"""
Tests for utility methods for serialization

"""

import numpy as np
import pytest
from openff.utilities import requires_package as requires_pkg

from openff.toolkit.utils.serialization import Serializable
from openff.toolkit.utils.utils import deserialize_numpy, serialize_numpy


class Thing(Serializable):
    def __init__(self, description, mylist):
        self.description = description
        self.mylist = mylist

    def to_dict(self):
        return {"description": self.description, "mylist": self.mylist}

    @classmethod
    def from_dict(cls, d):
        return cls(d["description"], d["mylist"])

    def __eq__(self, other):
        """Comparator for asserting object field equality."""
        equality = self.description == other.description
        equality &= self.mylist == other.mylist
        return equality


# DEBUG
def write(filename, contents):
    if type(contents) is str:
        mode = "w"
    elif type(contents) is bytes:
        mode = "wb"
    else:
        raise Exception(f"Cannot handle contents of type {type(contents)}")
    with open(filename, mode) as outfile:
        outfile.write(contents)


class TestUtilsSerialization:
    """Test serialization and deserialization of a simple class."""

    @classmethod
    def setup_class(cls):
        cls.thing = Thing("blorb", [1, 2, 3])

    def test_json(self):
        """Test JSON serialization"""
        json_thing = self.thing.to_json()
        thing_from_json = self.thing.__class__.from_json(json_thing)
        assert self.thing == thing_from_json

    @requires_pkg("yaml")
    def test_yaml(self):
        """Test YAML serialization"""
        yaml_thing = self.thing.to_yaml()
        thing_from_yaml = self.thing.__class__.from_yaml(yaml_thing)
        assert self.thing == thing_from_yaml

    @requires_pkg("bson")
    def test_bson(self):
        """Test BSON serialization"""
        bson_thing = self.thing.to_bson()
        thing_from_bson = self.thing.__class__.from_bson(bson_thing)
        assert self.thing == thing_from_bson

    @pytest.mark.wip(
        reason=(
            "the current implementation of to_toml cannot handle dict "
            "keys associated to None (e.g. the ToolkitAM1BCC tag in the "
            "TestUtilsSMIRNOFFSerialization suite)."
        )
    )
    @requires_pkg("toml")
    def test_toml(self):
        """Test TOML serialization"""
        toml_thing = self.thing.to_toml()
        thing_from_toml = self.thing.__class__.from_toml(toml_thing)
        assert self.thing == thing_from_toml

    @requires_pkg("msgpack")
    def test_messagepack(self):
        """Test MessagePack serialization"""
        messagepack_thing = self.thing.to_messagepack()
        thing_from_messagepack = self.thing.__class__.from_messagepack(messagepack_thing)
        assert self.thing == thing_from_messagepack

    @pytest.mark.wip(
        reason="from/to_xml is not implemented yet. This test fails "
        "because the list of integers is saved in the XML as "
        "a list of numeric strings."
    )
    def test_xml(self):
        """Test XML serialization"""
        xml_thing = self.thing.to_xml()
        thing_from_xml = self.thing.__class__.from_xml(xml_thing)
        assert self.thing == thing_from_xml

    def test_pickle(self):
        """Test pickle serialization"""
        pkl_thing = self.thing.to_pickle()
        thing_from_pkl = self.thing.__class__.from_pickle(pkl_thing)
        assert self.thing == thing_from_pkl


class TestNumPySerialization:
    @pytest.mark.parametrize("endian", [">", "<"])
    def test_serialize_endianness(self, endian):
        """Test that arrays are serialized as big-endian"""
        # Force big- (">") or little-endianness ("<") , overriding architecture default
        dtype = np.dtype(float).newbyteorder(endian)
        arr = np.arange(3).astype(dtype)

        # This bytestring should be a big-endian representation, generated by
        # np.arange(3).astype(np.dtype(float).newbyteorder(">")).tobytes()
        assert serialize_numpy(arr)[0] == (
            b"\x00\x00\x00\x00\x00\x00\x00\x00?\xf0\x00\x00\x00\x00\x00\x00@\x00\x00\x00\x00\x00\x00\x00"
        )

    @pytest.mark.parametrize("endian", [">", "<"])
    def test_deserialize_endianness(self, endian):
        """Test that arrays are deserialized as big-endian, and that
        the deserialization breaks if the input is little-endian."""
        dt_input = np.dtype(float).newbyteorder(endian)
        arr = np.arange(3, dtype=dt_input)
        deserialized = deserialize_numpy(arr.tobytes(), arr.shape)
        assert np.allclose(arr, deserialized) == (endian == ">")


class DictionaryContainer(Serializable):
    def __init__(self, dictionary):
        import copy

        self.dictionary = copy.deepcopy(dictionary)

    def to_dict(self):
        return self.dictionary

    @staticmethod
    def from_dict(dictionary):
        return DictionaryContainer(dictionary)

    def __eq__(self, other):
        """Comparator for asserting object field equality."""
        return self.dictionary == other.dictionary


class TestUtilsSMIRNOFFSerialization(TestUtilsSerialization):
    """Test serialization and deserialization of smirnoff99Frosst."""

    @classmethod
    def setup_class(cls):
        cls.thing = Thing("blorb", [1, 2, 3])
        # Create an example object holding the SMIRNOFF xmltodict dictionary representation
        import xmltodict

        from openff.toolkit.utils import get_data_file_path

        filename = get_data_file_path("test_forcefields/test_forcefield.offxml")
        with open(filename) as f:
            xml = f.read()
            dictionary = xmltodict.parse(xml)
            cls.thing = DictionaryContainer(dictionary)



"Registry for ToolkitWrapper objects"

__all__ = ("ToolkitRegistry", "toolkit_registry_manager")

import inspect
import logging
from collections.abc import Callable
from contextlib import contextmanager

from openff.toolkit.utils.ambertools_wrapper import AmberToolsToolkitWrapper
from openff.toolkit.utils.base_wrapper import ToolkitWrapper
from openff.toolkit.utils.builtin_wrapper import BuiltInToolkitWrapper
from openff.toolkit.utils.exceptions import (
    InvalidToolkitError,
    LicenseError,
    ToolkitUnavailableException,
)
from openff.toolkit.utils.openeye_wrapper import OpenEyeToolkitWrapper
from openff.toolkit.utils.rdkit_wrapper import RDKitToolkitWrapper
from openff.toolkit.utils.utils import all_subclasses

logger = logging.getLogger(__name__)


class ToolkitRegistry:
    """
    A registry and precedence list for :py:class:`ToolkitWrapper` objects.

    ``ToolkitRegistry`` allows the OpenFF Toolkit to provide a concise,
    universal API that can call out to other well-established libraries rather
    than re-implement algorithms with well-established implementations, while
    still giving users control over dependencies. It contains a list of
    :py:class:`ToolkitWrapper` objects, each of which provides some collection
    of methods that may be requested from the registry. The :py:meth:`call`
    method takes a name and calls the method of that name on each wrapper until
    it finds a working implementation, whose result it returns. For details on
    how this search is conducted and what counts as a working implementation,
    see that method's API docs.

    Examples
    --------

    Register toolkits in a specified order, skipping if unavailable

    >>> from openff.toolkit.utils.toolkits import ToolkitRegistry
    >>> toolkit_precedence = [OpenEyeToolkitWrapper, RDKitToolkitWrapper, AmberToolsToolkitWrapper]
    >>> toolkit_registry = ToolkitRegistry(toolkit_precedence)
    >>> toolkit_registry
    <ToolkitRegistry containing OpenEye Toolkit, The RDKit, AmberTools>

    Register all available toolkits (in the order OpenEye, RDKit, AmberTools,
    built-in)

    >>> toolkits = [OpenEyeToolkitWrapper, RDKitToolkitWrapper, AmberToolsToolkitWrapper, BuiltInToolkitWrapper]
    >>> toolkit_registry = ToolkitRegistry(toolkit_precedence=toolkits)
    >>> toolkit_registry
    <ToolkitRegistry containing OpenEye Toolkit, The RDKit, AmberTools, Built-in Toolkit>

    Retrieve the global singleton toolkit registry, which is created when this
    module is imported from all available toolkits:

    >>> from openff.toolkit.utils.toolkits import GLOBAL_TOOLKIT_REGISTRY
    >>> GLOBAL_TOOLKIT_REGISTRY
    <ToolkitRegistry containing OpenEye Toolkit, The RDKit, AmberTools, Built-in Toolkit>

    Note that this will contain different ToolkitWrapper objects based on what
    toolkits are currently installed.

    Call a method on the registered toolkit wrapper with the highest precedence:

    >>> molecule = toolkit_registry.call('from_smiles', 'Cc1ccccc1')

    For more, see the :py:meth:`call` method.

    .. warning :: This API is experimental and subject to change.
    """

    _toolkits: list[ToolkitWrapper]

    def __init__(
        self,
        toolkit_precedence: list[type[ToolkitWrapper]] | None = None,
        exception_if_unavailable: bool = True,
        _register_imported_toolkit_wrappers: bool = False,
    ):
        """
        Create an empty toolkit registry.

        Parameters
        ----------
        toolkit_precedence
            List of toolkit wrapper classes, in order of desired precedence when performing molecule operations. If
            None, no toolkits will be registered.

        exception_if_unavailable
            If True, an exception will be raised if the toolkit is unavailable

        _register_imported_toolkit_wrappers
            If True, will attempt to register all imported ToolkitWrapper subclasses that can be
            found in the order of toolkit_precedence, if specified. If toolkit_precedence is not
            specified, the default order is [OpenEyeToolkitWrapper, RDKitToolkitWrapper,
            AmberToolsToolkitWrapper, BuiltInToolkitWrapper].

        """
        self._toolkits = list()

        toolkit_classes_to_register: list[type[ToolkitWrapper]] = list()

        if _register_imported_toolkit_wrappers:
            if toolkit_precedence is None:
                toolkit_precedence = [
                    OpenEyeToolkitWrapper,
                    RDKitToolkitWrapper,
                    AmberToolsToolkitWrapper,
                    BuiltInToolkitWrapper,
                ]
            all_importable_toolkit_wrappers = all_subclasses(ToolkitWrapper)
            for toolkit in toolkit_precedence:
                if toolkit in all_importable_toolkit_wrappers:
                    toolkit_classes_to_register.append(toolkit)
        else:
            if toolkit_precedence is not None:
                toolkit_classes_to_register = toolkit_precedence

        if toolkit_classes_to_register:
            for toolkit in toolkit_classes_to_register:
                self.register_toolkit(
                    toolkit_wrapper=toolkit,
                    exception_if_unavailable=exception_if_unavailable,
                )

    @property
    def registered_toolkits(self) -> list[ToolkitWrapper]:
        """
        List registered toolkits.

        .. warning :: This API is experimental and subject to change.

        .. todo :: Should this return a generator? Deep copies? Classes? Toolkit names?

        Returns
        -------
        toolkits
        """
        return list(self._toolkits)

    @property
    def registered_toolkit_versions(self) -> dict[str, str]:
        """
        Return a dict containing the version of each registered toolkit.

        .. warning :: This API is experimental and subject to change.

        Returns
        -------
        toolkit_versions
            A dictionary mapping names and versions of wrapped toolkits

        """
        return {tk.toolkit_name: tk.toolkit_version for tk in self.registered_toolkits}

    # TODO: Can we instead register available methods directly with `ToolkitRegistry`,
    # so we can just use `ToolkitRegistry.method()`?
    def call(
        self,
        method_name: str,
        *args,
        raise_exception_types: list[type[Exception]] | None = None,
        **kwargs,
    ):
        """
        Execute a method with the first registered toolkits that supports it.

        This method searches the registry's precedence list for the first
        wrapper that has an attribute named ``method_name`` and attempts to
        call it as a method using the arguments in ``*args`` and ``**kwargs``.
        If that method raises no exception, its return value is returned.

        By default, if a wrapper with an appropriately-named method raises an
        exception of any type, then iteration over the registered toolkits
        stops early and that exception is raised. To limit this behavior to only
        certain exceptions and otherwise continue iteration, customize this
        behavior using the optional ``raise_exception_types`` keyword argument.
        If iteration finishes without finding a wrapper that can successfully
        call the requested method, a ``ValueError`` is raised, containing a
        message listing the registered toolkits and any exceptions that
        occurred.

        Parameters
        ----------
        method_name
            The name of the method to execute
        raise_exception_types
            A list of exception-derived types to catch and raise immediately. If
            ``None``, the first exception encountered will be raised. To ignore
            all exceptions, set this to the empty list ``[]``.

        Raises
        ------
        ValueError
            If no suitable toolkit wrapper was found in the registry.

        Exception
            Other forms of exceptions are possible if raise_exception_types is
            specified. These are defined by the ``ToolkitWrapper`` method being
            called.

        Examples
        --------

        Create a molecule, and call the toolkit ``to_smiles()`` method directly

        >>> from openff.toolkit import (
        ...     Molecule,
        ...     ToolkitRegistry,
        ...     AmberToolsToolkitWrapper,
        ...     RDKitToolkitWrapper,
        ... )
        >>> molecule = Molecule.from_smiles('Cc1ccccc1')
        >>> toolkit_registry = ToolkitRegistry([
        ...     AmberToolsToolkitWrapper,
        ...     RDKitToolkitWrapper,
        ... ])
        >>> smiles = toolkit_registry.call('to_smiles', molecule)

        Stop if a partial charge assignment method encounters an error during
        the partial charge calculation (:py:exc:`ChargeCalculationError`), but
        proceed to the next wrapper for any other exception such as the wrapper
        not supporting the charge method:

        >>> from openff.toolkit import GLOBAL_TOOLKIT_REGISTRY
        >>> from openff.toolkit.utils.exceptions import ChargeCalculationError
        >>> GLOBAL_TOOLKIT_REGISTRY.call(
        ...     "assign_partial_charges",
        ...     molecule=Molecule.from_smiles('C'),
        ...     partial_charge_method="gasteiger",
        ...     raise_exception_types=[ChargeCalculationError],
        ... )

        Calling a method that exists on none of the wrappers raises a
        ``ValueError``:

        >>> from openff.toolkit import ToolkitRegistry, RDKitToolkitWrapper
        >>> toolkit_registry = ToolkitRegistry([RDKitToolkitWrapper])
        >>> toolkit_registry.call("there_is_no_spoon") # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
        Traceback (most recent call last):
          ...
        ValueError: No registered toolkits can provide the capability
        "there_is_no_spoon" for args "()" and kwargs "{}"
        Available toolkits are: [ToolkitWrapper around The RDKit version ...]
        <BLANKLINE>

        """
        if raise_exception_types is None:
            raise_exception_types = [Exception]

        errors = list()
        for toolkit in self._toolkits:
            if hasattr(toolkit, method_name):
                method = getattr(toolkit, method_name)
                try:
                    return method(*args, **kwargs)
                except Exception as e:
                    for exception_type in raise_exception_types:
                        if isinstance(e, exception_type):
                            raise e
                    errors.append((toolkit, e))

        # No toolkit was found to provide the requested capability
        # TODO: Can we help developers by providing a check for typos in expected method names?
        msg = (
            f'No registered toolkits can provide the capability "{method_name}" '
            f'for args "{args}" and kwargs "{kwargs}"\n'
        )

        msg += f"Available toolkits are: {self.registered_toolkits}\n"
        # Append information about toolkits that implemented the method, but could not handle the provided parameters
        for toolkit, error in errors:
            msg += f" {toolkit} {type(error)} : {error}\n"
        raise ValueError(msg)

    def __repr__(self):
        return f"<ToolkitRegistry containing {', '.join([tk.toolkit_name for tk in self._toolkits])}>"

    def register_toolkit(
        self,
        toolkit_wrapper: ToolkitWrapper | type[ToolkitWrapper],
        exception_if_unavailable: bool = True,
    ):
        """
        Register the provided toolkit wrapper class, instantiating an object of it.

        .. warning :: This API is experimental and subject to change.

        .. todo ::

           This method should raise an exception if the toolkit is unavailable, unless an optional argument
           is specified that silently avoids registration of toolkits that are unavailable.

        Parameters
        ----------
        toolkit_wrapper
            The toolkit wrapper to register or its class.
        exception_if_unavailable
            If True, an exception will be raised if the toolkit is unavailable

        """
        # Instantiate class if class, or just add if already instantiated.
        if isinstance(toolkit_wrapper, ToolkitWrapper):
            self._toolkits.append(toolkit_wrapper)

        elif issubclass(toolkit_wrapper, ToolkitWrapper):
            try:
                _toolkit_wrapper = toolkit_wrapper()

            # This exception can be raised by OpenEyeToolkitWrapper
            except LicenseError as license_exception:
                if exception_if_unavailable:
                    raise ToolkitUnavailableException(license_exception.msg)
                else:
                    logger.warning(license_exception)
                return
            except ToolkitUnavailableException:
                if exception_if_unavailable:
                    raise ToolkitUnavailableException(f"Unable to load toolkit '{toolkit_wrapper}'. ")
                return

            self._toolkits.append(_toolkit_wrapper)

        else:
            raise ValueError(f"Given unexpected argument {type(toolkit_wrapper)=}")

    def deregister_toolkit(self, toolkit_wrapper: ToolkitWrapper):
        """
        Remove a ToolkitWrapper from the list of toolkits in this ToolkitRegistry

        .. warning :: This API is experimental and subject to change.

        Parameters
        ----------
        toolkit_wrapper
            The toolkit wrapper to remove from the registry

        Raises
        ------
        InvalidToolkitError
            If toolkit_wrapper is not a ToolkitWrapper or subclass
        ToolkitUnavailableException
            If toolkit_wrapper is not found in the registry
        """
        # If passed a class, instantiate it
        if inspect.isclass(toolkit_wrapper):
            toolkit_wrapper = toolkit_wrapper()

        if not isinstance(toolkit_wrapper, ToolkitWrapper):
            msg = (
                f"Argument {toolkit_wrapper} must an ToolkitWrapper "
                f"or subclass of it. Found type {type(toolkit_wrapper)}."
            )
            raise InvalidToolkitError(msg)

        toolkits_to_remove = []

        for toolkit in self._toolkits:
            if type(toolkit) is type(toolkit_wrapper):
                toolkits_to_remove.append(toolkit)

        if not toolkits_to_remove:
            msg = f"Did not find {toolkit_wrapper} in registry. Currently registered toolkits are {self._toolkits}"
            raise ToolkitUnavailableException(msg)

        for toolkit_to_remove in toolkits_to_remove:
            self._toolkits.remove(toolkit_to_remove)

    def add_toolkit(self, toolkit_wrapper: ToolkitWrapper):
        """
        Append a ToolkitWrapper onto the list of toolkits in this ToolkitRegistry

        .. warning :: This API is experimental and subject to change.

        Parameters
        ----------
        toolkit_wrapper
            The ToolkitWrapper object to add to the list of registered toolkits

        Raises
        ------
        InvalidToolkitError
            If toolkit_wrapper is not a ToolkitWrapper or subclass
        """
        if not isinstance(toolkit_wrapper, ToolkitWrapper):
            raise InvalidToolkitError(
                "Something other than a ToolkitWrapper object was passed to ToolkitRegistry.add_toolkit()\n"
                f"Given object {toolkit_wrapper} of type {type(toolkit_wrapper)}"
            )

        self._toolkits.append(toolkit_wrapper)

    # TODO: Can we automatically resolve calls to methods that are not explicitly defined using some Python magic?

    def resolve(self, method_name: str) -> Callable:
        """
        Get a method with the given name from the first registered toolkit that provides it.

        Resolve the requested method name by checking all registered toolkits in
        order of precedence for one that provides the requested method. Note
        that this may not be the method used by :py:meth:`call` if the
        ``raise_exception_types`` argument is passed.


        Parameters
        ----------
        method_name
            The name of the method to resolve

        Returns
        -------
        method
            The method of the first registered toolkit that provides the
            requested method name

        Raises
        ------
        NotImplementedError
            if the requested method cannot be found among the registered toolkits

        Examples
        --------

        Create a molecule, and call the toolkit ``to_smiles()`` method directly

        >>> from openff.toolkit import Molecule
        >>> molecule = Molecule.from_smiles('Cc1ccccc1')
        >>> toolkit_registry = ToolkitRegistry([
        ...     OpenEyeToolkitWrapper,
        ...     RDKitToolkitWrapper,
        ...     AmberToolsToolkitWrapper
        ... ])
        >>> method = toolkit_registry.resolve('to_smiles')
        >>> smiles = method(molecule)

        .. todo :: Is there a better way to figure out which toolkits implement given methods by introspection?

        """
        for toolkit in self._toolkits:
            if hasattr(toolkit, method_name):
                method = getattr(toolkit, method_name)
                return method

        # No toolkit was found to provide the requested capability
        # TODO: Can we help developers by providing a check for typos in expected method names?
        raise NotImplementedError(
            f'No registered toolkits can provide the capability "{method_name}".\n'
            f"Available toolkits are: {self.registered_toolkits}\n"
        )


# Copied from https://github.com/openforcefield/openff-fragmenter/blob/4a290b866a8ed43eabcbd3231c62b01f0c6d7df6
# /openff/fragmenter/utils.py#L97-L123
@contextmanager
def toolkit_registry_manager(toolkit_registry: ToolkitRegistry | ToolkitWrapper):
    """
    A context manager that temporarily changes the :py:data:`GLOBAL_TOOLKIT_REGISTRY`.

    This can be useful in cases where one would otherwise need to otherwise
    manually specify the use of a specific :py:class:`ToolkitWrapper` or
    :py:class:`ToolkitRegistry` repeatedly in a block of code, or in cases where
    there isn't another way to switch the ``ToolkitWrapper`` used for a
    particular operation.

    Examples
    --------
    >>> from openff.toolkit import Molecule, RDKitToolkitWrapper, AmberToolsToolkitWrapper
    >>> from openff.toolkit.utils import toolkit_registry_manager, ToolkitRegistry
    >>> mol = Molecule.from_smiles("CCO")
    >>> print(mol.to_smiles()) # This will use the OpenEye backend (if installed and licensed)
    [H]C([H])([H])C([H])([H])O[H]
    >>> with toolkit_registry_manager(ToolkitRegistry([RDKitToolkitWrapper()])):
    ...    print(mol.to_smiles())
    [H][O][C]([H])([H])[C]([H])([H])[H]

    """
    from openff.toolkit.utils.toolkits import GLOBAL_TOOLKIT_REGISTRY

    if isinstance(toolkit_registry, ToolkitRegistry):
        context_toolkits = toolkit_registry.registered_toolkits
    elif isinstance(toolkit_registry, ToolkitWrapper):
        context_toolkits = [toolkit_registry]
    else:
        raise NotImplementedError("Only ``ToolkitRegistry`` and ``ToolkitWrapper`` are supported.")

    original_toolkits = GLOBAL_TOOLKIT_REGISTRY.registered_toolkits

    for toolkit in original_toolkits:
        GLOBAL_TOOLKIT_REGISTRY.deregister_toolkit(toolkit)

    for toolkit in context_toolkits:
        GLOBAL_TOOLKIT_REGISTRY.register_toolkit(toolkit)

    try:
        yield
    finally:
        for toolkit in context_toolkits:
            GLOBAL_TOOLKIT_REGISTRY.deregister_toolkit(toolkit)
        for toolkit in original_toolkits:
            GLOBAL_TOOLKIT_REGISTRY.register_toolkit(toolkit)


_toolkit_registry_manager = toolkit_registry_manager



import uuid
from io import StringIO
from typing import TYPE_CHECKING

from nglview.base_adaptor import Structure, Trajectory

if TYPE_CHECKING:
    from openff.toolkit import Molecule, Topology

MOLECULE_DEFAULT_REPS = [dict(type="licorice", params=dict(radius=0.25, multipleBond=True))]


class MoleculeNGLViewTrajectory(Structure, Trajectory):
    """
    OpenFF Molecule adaptor.

    The file format used to communicate with NGLView can make a big difference
    in the visualization. For instance, using PDB will require NGLView to infer
    bonds from positions, whereas using MOL2 or SDF may result in a loss of
    residue data.

    Parameters
    ----------
    molecule
        The `Molecule` object to display.
    ext
        The file extension to use to communicate with NGLView. The format must
        be supported for export by the Toolkit via the `Molecule.to_file()
        <openff.toolkit.topology.molecule.to_file>` method, and import by
        NGLView. File formats supported by NGLView can be found at
        https://nglviewer.org/ngl/api/manual/file-formats.html

    Example
    -------
    >>> import nglview as nv
    >>>
    >>> mol = Molecule.from_polymer_pdb("file.pdb")  # doctest: +SKIP
    >>> nv.NGLWidget(MoleculeNGLViewTrajectory(mol))  # doctest: +SKIP
    """

    def __init__(
        self,
        molecule: "Molecule",
        ext: str = "MOL2",
    ):
        if not molecule.conformers:
            raise ValueError("Cannot visualize a molecule without conformers with NGLView")
        self.molecule = molecule
        self.ext = ext.lower()
        self.params: dict = dict()
        self.id = str(uuid.uuid4())

    def get_coordinates(self, index: int = 0):
        return self.molecule.conformers[index].m_as("angstrom")

    @property
    def n_frames(self):
        return len(self.molecule.conformers)

    def get_structure_string(self):
        with StringIO() as f:
            self.molecule.to_file(f, file_format=self.ext)
            structure_string = f.getvalue()
        return structure_string


class TopologyNGLViewStructure(Structure):
    """
    OpenFF Topology adaptor.

    Parameters
    ----------
    topology
        The `Topology` object to display.
    ext
        The file extension to use to communicate with NGLView. The format must
        be supported for export by the Toolkit via the `Topology.to_file()
        <openff.toolkit.topology.Topology.to_file>` method, and import by
        NGLView. File formats supported by NGLView can be found at
        https://nglviewer.org/ngl/api/manual/file-formats.html

    Example
    -------
    >>> import nglview as nv
    >>>
    >>> top = Topology.from_pdb("file.pdb")  # doctest: +SKIP
    >>> nv.NGLWidget(TopologyNGLViewStructure(top))  # doctest: +SKIP
    """

    def __init__(
        self,
        topology: "Topology",
        ext: str = "PDB",
    ):
        self.topology = topology
        self.ext = ext.lower()
        self.params: dict = dict()
        self.id = str(uuid.uuid4())

    def get_structure_string(self):
        with StringIO() as f:
            self.topology.to_file(f, file_format=self.ext)
            structure_string = f.getvalue()
        return structure_string



from openff.toolkit.utils.constants import (
    ALLOWED_AROMATICITY_MODELS,
    ALLOWED_CHARGE_MODELS,
    ALLOWED_FRACTIONAL_BOND_ORDER_MODELS,
    DEFAULT_AROMATICITY_MODEL,
    DEFAULT_CHARGE_MODEL,
    DEFAULT_FRACTIONAL_BOND_ORDER_MODEL,
)
from openff.toolkit.utils.toolkit_registry import toolkit_registry_manager
from openff.toolkit.utils.toolkits import (
    AMBERTOOLS_AVAILABLE,
    BASIC_CHEMINFORMATICS_TOOLKITS,
    GLOBAL_TOOLKIT_REGISTRY,
    OPENEYE_AVAILABLE,
    RDKIT_AVAILABLE,
    AmberToolsToolkitWrapper,
    AntechamberNotFoundError,
    BadMoleculeAssumptionError,
    BuiltInToolkitWrapper,
    ChargeCalculationError,
    ChargeMethodUnavailableError,
    GAFFAtomTypeWarning,
    IncompatibleUnitError,
    IncorrectNumConformersError,
    IncorrectNumConformersWarning,
    InvalidIUPACNameError,
    InvalidToolkitError,
    InvalidToolkitRegistryError,
    LicenseError,
    MissingPackageError,
    OpenEyeToolkitWrapper,
    RDKitToolkitWrapper,
    SMILESParseError,
    ToolkitRegistry,
    ToolkitUnavailableException,
    ToolkitWrapper,
    UndefinedStereochemistryError,
)
from openff.toolkit.utils.utils import (
    all_subclasses,
    convert_0_1_smirnoff_to_0_2,
    convert_0_2_smirnoff_to_0_3,
    convert_all_quantities_to_string,
    convert_all_strings_to_quantity,
    deserialize_numpy,
    get_data_file_path,
    get_molecule_parameterIDs,
    inherit_docstrings,
    object_to_quantity,
    quantity_to_string,
    requires_package,
    serialize_numpy,
    string_to_quantity,
    string_to_unit,
    temporary_cd,
    unit_to_string,
)



"""
Parameter assignment tools for the SMIRNOFF (SMIRKS Native Open Force Field) format.

.. codeauthor:: John D. Chodera <john.chodera@choderalab.org>
.. codeauthor:: David L. Mobley <dmobley@mobleylab.org>
.. codeauthor:: Peter K. Eastman <peastman@stanford.edu>

.. todo ::

   * Speed up overall import time by putting non-global imports only where they are needed

"""

import logging
import os
import pathlib
from importlib.metadata import entry_points
from typing import IO, TYPE_CHECKING, Any, Union

from packaging.version import Version

from openff.toolkit import Quantity
from openff.toolkit.typing.engines.smirnoff.io import ParameterIOHandler
from openff.toolkit.typing.engines.smirnoff.parameters import ParameterHandler
from openff.toolkit.typing.engines.smirnoff.plugins import load_handler_plugins
from openff.toolkit.utils.constants import (
    ALLOWED_AROMATICITY_MODELS,
    DEFAULT_AROMATICITY_MODEL,
)
from openff.toolkit.utils.exceptions import (
    ParameterHandlerRegistrationError,
    PartialChargeVirtualSitesError,
    SMIRNOFFAromaticityError,
    SMIRNOFFParseError,
    SMIRNOFFVersionError,
)
from openff.toolkit.utils.utils import (
    all_subclasses,
    convert_0_1_smirnoff_to_0_2,
    convert_0_2_smirnoff_to_0_3,
    convert_all_quantities_to_string,
    convert_all_strings_to_quantity,
    requires_package,
)

if TYPE_CHECKING:
    import openmm
    from openff.interchange import Interchange

    from openff.toolkit import Molecule
    from openff.toolkit.topology.topology import Topology, ValenceDict
    from openff.toolkit.utils.base_wrapper import ToolkitWrapper
    from openff.toolkit.utils.toolkit_registry import ToolkitRegistry


__all__ = [
    "MAX_SUPPORTED_VERSION",
    "ForceField",
    "ParameterHandlerRegistrationError",
    "PartialChargeVirtualSitesError",
    "SMIRNOFFAromaticityError",
    "SMIRNOFFParseError",
    "SMIRNOFFVersionError",
    "get_available_force_fields",
]

logger = logging.getLogger(__name__)

# Directory paths used by ForceField to discover offxml files.
_installed_offxml_dir_paths: list[str] = []


def _get_installed_offxml_dir_paths() -> list[str]:
    """Return the list of directory paths where to search for offxml files.

    This function load the information by calling all the entry points
    registered in the "openff.forcefielddirs" group. Each entry point
    (i.e. a function) should return a list of paths to directories
    containing the offxml files.

    Returns
    -------
    installed_offxml_dir_paths :list[str]
        All the installed directory paths where ``ForceField`` will
        look for offxml files.

    """
    global _installed_offxml_dir_paths
    if len(_installed_offxml_dir_paths) == 0:
        # Find all registered entry points that should return a list of
        # paths to directories where to search for offxml files.
        for entry_point in entry_points().select(group="openforcefield.smirnoff_forcefield_directory"):
            _installed_offxml_dir_paths.extend(entry_point.load()())
    return _installed_offxml_dir_paths


def get_available_force_fields(full_paths=False):
    r"""
    Get the filenames of all available .offxml force field files.

    Availability is determined by what is discovered through the
    ``openforcefield.smirnoff_forcefield_directory`` entry point. If the
    ``openff-forcefields`` package is installed, this should include several
    .offxml files such as ``openff-1.0.0.offxml``.

    Parameters
    ----------
    full_paths
        If False, return the name of each available \*.offxml file.
        If True, return the full path to each available \*.offxml file.

    Returns
    -------
    available_force_fields
        List of available force field files

    """
    installed_paths = _get_installed_offxml_dir_paths()
    available_force_fields = []
    for installed_path in installed_paths:
        for globbed in pathlib.Path(installed_path).rglob("*.offxml"):
            if full_paths:
                available_force_fields.append(globbed.as_posix())
            else:
                available_force_fields.append(globbed.name)
    return available_force_fields


# TODO: Instead of having a global version number, alow each Force to have a separate version number
MAX_SUPPORTED_VERSION = "1.0"  # maximum version of the SMIRNOFF spec supported by this SMIRNOFF force field


class ForceField:
    """
    A factory that assigns SMIRNOFF parameters to a molecular system

    :class:`ForceField` is a factory that constructs an
    OpenMM :class:`System <openmm.openmm.System>` object from a
    :class:`Topology <openff.toolkit.topology.Topology>` object defining a
    (bio)molecular system containing one or more molecules.

    When a :class:`ForceField` object is created from one or more specified
    SMIRNOFF serialized representations, all :class:`ParameterHandler
    <openff.toolkit.typing.engines.smirnoff.parameters.ParameterHandler>`
    subclasses currently imported are identified and registered to handle
    different sections of the SMIRNOFF force field definition file(s).

    All :class:`ParameterIOHandler` subclasses currently imported are identified
    and registered to handle different serialization formats (such as XML).

    The force field definition is processed by these handlers to populate the
    ``ForceField`` object model data structures that can easily be manipulated
    via the API:

    Processing a :class:`Topology` object defining a chemical system will then
    call all :class:`ParameterHandler` objects in an order guaranteed to
    satisfy the declared processing order constraints of each
    :class:`ParameterHandler`.

    Examples
    --------

    Create a new ForceField object from the distributed OpenFF 2.0 ("Sage") file:

    >>> from openff.toolkit import ForceField
    >>> force_field = ForceField('openff-2.0.0.offxml')

    Create an OpenMM system from a :class:`openff.toolkit.topology.Topology` object:

    >>> from openff.toolkit import Molecule, Topology
    >>> ethanol = Molecule.from_smiles('CCO')
    >>> topology = Topology.from_molecules(molecules=[ethanol])
    >>> system = force_field.create_openmm_system(topology)

    Modify the long-range electrostatics method:

    >>> force_field.get_parameter_handler('Electrostatics').periodic_potential = 'PME'

    Inspect the first few vdW parameters:

    >>> low_precedence_parameters = force_field.get_parameter_handler('vdW').parameters[0:3]

    Retrieve the vdW parameters by SMIRKS string and manipulate it:

    >>> from openff.toolkit import unit
    >>> parameter = force_field.get_parameter_handler('vdW').parameters['[#1:1]-[#7]']
    >>> parameter.rmin_half += 0.1 * unit.angstroms
    >>> parameter.epsilon *= 1.02

    Make a child vdW type more specific (checking modified SMIRKS for validity):

    >>> force_field.get_parameter_handler('vdW').parameters[-1].smirks += '~[#53]'

    .. warning ::

       While we check whether the modified SMIRKS is still valid and has the appropriate valence type,
       we currently don't check whether the typing remains hierarchical, which could result in some types
       no longer being assignable because more general types now come *below* them and preferentially match.

    Delete a parameter:

    >>> del force_field.get_parameter_handler('vdW').parameters['[#1:1]-[#6X4]']

    Insert a parameter at a specific point in the parameter tree:

    >>> from openff.toolkit.typing.engines.smirnoff import vdWHandler
    >>> new_parameter = vdWHandler.vdWType(
    ...     smirks='[*:1]',
    ...     epsilon=0.0157*unit.kilocalories_per_mole,
    ...     rmin_half=0.6000*unit.angstroms,
    ... )
    >>> force_field.get_parameter_handler('vdW').parameters.insert(0, new_parameter)

    .. warning ::

       We currently don't check whether removing a parameter could accidentally remove the root type, so it's possible
       to no longer type all molecules this way.

    """

    def __init__(
        self,
        *sources,
        aromaticity_model: str = DEFAULT_AROMATICITY_MODEL,
        parameter_handler_classes: list[type[ParameterHandler]] | None = None,
        parameter_io_handler_classes: list[type[ParameterIOHandler]] | None = None,
        disable_version_check: bool = False,
        allow_cosmetic_attributes: bool = False,
        load_plugins: bool = False,
    ):
        """Create a new :class:`ForceField` object from one or more SMIRNOFF parameter definition files.

        Parameters
        ----------
        sources
            A list of files defining the SMIRNOFF force field to be loaded.
            Currently, only `the SMIRNOFF XML format <https://openforcefield.github.io/standards/standards/smirnoff/>`_
            is supported.  Each entry may be an absolute file path, a path relative to the current working directory, a
            path relative to this module's data subdirectory (for built in force fields), or an open file-like object
            with a ``read()`` method from which the force field XML data can be loaded.  If multiple files are
            specified, any top-level tags that are repeated will be merged if they are compatible, with files appearing
            later in the sequence resulting in parameters that have higher precedence.  Support for multiple files is
            primarily intended to allow solvent parameters to be specified by listing them last in the sequence.
        aromaticity_model
            The aromaticity model to use. Only OEAroModel_MDL is supported.
        parameter_handler_classes
            If not None, the specified set of ParameterHandler classes will be instantiated to create the parameter
            object model.  By default, all imported subclasses of ParameterHandler not loaded as plugins are
            automatically registered.
        parameter_io_handler_classes
            If not None, the specified set of ParameterIOHandler classes will be used to parse/generate serialized
            parameter sets.  By default, all imported subclasses of ParameterIOHandler are automatically registered.
        disable_version_check
            If True, will disable checks against the current highest supported force field version.
            This option is primarily intended for force field development.
        allow_cosmetic_attributes
            Whether to retain non-spec kwargs from data sources.
        load_plugins
            Whether to load ``ParameterHandler`` classes which have been registered
            by installed plugins.

        Examples
        --------

        Load one SMIRNOFF parameter set in XML format (searching the package data directory by default, which includes
        some standard parameter sets):

        >>> forcefield = ForceField('openff-2.0.0.offxml')

        Load multiple SMIRNOFF parameter sets:

        >>> from openff.toolkit._tests.utils import get_data_file_path
        >>> forcefield = ForceField('openff-2.0.0.offxml', get_data_file_path('test_forcefields/tip3p.offxml'))

        Load a parameter set from a string:

        >>> offxml = '<SMIRNOFF version="0.2" aromaticity_model="OEAroModel_MDL"/>'
        >>> forcefield = ForceField(offxml)

        See Also
        --------
        parse_sources

        Notes
        -----
        No effort is made to de-duplicate redundant parameters or parameters with identical SMIRKS patterns.

        """
        # Clear all object fields
        self._initialize()

        self.aromaticity_model = aromaticity_model
        # Store initialization options
        self.disable_version_check = disable_version_check
        # if True, we won't check which SMIRNOFF version number we're parsing

        # Register all ParameterHandler objects that will process SMIRNOFF force definitions
        # TODO: We need to change this to just find all ParameterHandler objects in this file;
        # otherwise, we can't define two different ParameterHandler subclasses to compare for a new type of energy term
        # since both will try to register themselves for the same XML tag and an Exception will be raised.
        if parameter_handler_classes is None:
            # TODO: What is the point of this argument? The default behavior includes a lot of logic.
            #       Unaware of cases in which the argument is used.

            internal_module = "openff.toolkit.typing.engines.smirnoff.parameters"

            # Plugins, if already loaded, are also subclasses of ParameterHandler and therefore
            # show up in all_subclasses, but we don't want them included here. Checking the
            # module each come from is the safest way to know which are plugins; another
            # option is explicitly defining a set of default handlers

            default_handlers: list[type[ParameterHandler]] = [
                handler for handler in all_subclasses(ParameterHandler) if handler.__module__ == internal_module
            ]

            # We could also track a list of the plugins that were already loaded, but that's not necessary here
            # since we're only cleaning out the handlers that were already registered.
            parameter_handler_classes = default_handlers

        if load_plugins:
            plugin_classes = load_handler_plugins()

            for plugin_class in plugin_classes:
                parameter_handler_classes.append(plugin_class)
                self._plugin_parameter_handler_classes.append(plugin_class)

        self._register_parameter_handler_classes(parameter_handler_classes)

        # Register all ParameterIOHandler objects that will process serialized parameter representations
        if parameter_io_handler_classes is None:
            parameter_io_handler_classes = all_subclasses(ParameterIOHandler)

        self._register_parameter_io_handler_classes(parameter_io_handler_classes)

        # Parse all sources containing SMIRNOFF parameter definitions
        self.parse_sources(sources, allow_cosmetic_attributes=allow_cosmetic_attributes)

    def _initialize(self):
        """
        Initialize all object fields.
        """
        self._MIN_SUPPORTED_SMIRNOFF_VERSION = Version("0.1")
        self._MAX_SUPPORTED_SMIRNOFF_VERSION = Version("0.3")
        # if True, will disable checking compatibility version
        self._disable_version_check = False
        self._aromaticity_model = DEFAULT_AROMATICITY_MODEL
        # Parameter handler classes that _can_ be initialized if needed
        self._parameter_handler_classes: dict[str, type[ParameterHandler]] = dict()
        # ParameterHandler classes to be instantiated for each parameter type
        self._parameter_handlers: dict[str, ParameterHandler] = dict()
        # classes of ParameterHandlers that were registered via the plugin interface
        self._plugin_parameter_handler_classes: list[type[ParameterHandler]] = list()
        # ParameterIOHandler classes that _can_ be initialiazed if needed
        self._parameter_io_handler_classes: dict[str, type[ParameterIOHandler]] = dict()
        # ParameterIO classes to be used for each file type
        self._parameter_io_handlers: dict[str, ParameterIOHandler] = dict()
        # It'd be much more convenient for these to only be str, and just default to the empty string
        self._author: str | None = None
        self._date: str | None = None

    def _check_smirnoff_version_compatibility(self, version: str):
        """
        Raise a parsing exception if the given file version is incompatible with this ForceField class.

        Parameters
        ----------
        version
            The SMIRNOFF version being read.

        Raises
        ------
        SMIRNOFFVersionError
            If an incompatible version is passed in.

        """
        from packaging.version import parse

        # Use PEP-440 compliant version number comparison, if requested
        if self.disable_version_check:
            pass
        else:
            if (parse(str(version)) > parse(str(self._MAX_SUPPORTED_SMIRNOFF_VERSION))) or (
                parse(str(version)) < parse(str(self._MIN_SUPPORTED_SMIRNOFF_VERSION))
            ):
                raise SMIRNOFFVersionError(
                    f"SMIRNOFF offxml file was written with version {version}, but this version of "
                    f"ForceField only supports version {self._MIN_SUPPORTED_SMIRNOFF_VERSION} to "
                    f"version {self._MAX_SUPPORTED_SMIRNOFF_VERSION}"
                )

    @property
    def aromaticity_model(self) -> str:
        """Returns the aromaticity model for this ForceField object.

        Returns
        -------
        aromaticity_model
            The aromaticity model for this force field.
        """
        return self._aromaticity_model

    @aromaticity_model.setter
    def aromaticity_model(self, aromaticity_model: str):
        """
        Register that this force field is using an aromaticity model. Will check for
        compatibility with other aromaticity model(s) already in use.

        Parameters
        ----------
        aromaticity_model
            The aromaticity model to register.

        Raises
        ------
        SMIRNOFFAromaticityError
            If an incompatible aromaticity model is passed in.

        Notes
        -----
           * Currently, the only supported aromaticity model is 'OEAroModel_MDL'.

        """
        # Implement better logic here if we ever support another aromaticity model
        if aromaticity_model not in ALLOWED_AROMATICITY_MODELS:
            raise SMIRNOFFAromaticityError(
                f"Read aromaticity model {aromaticity_model} which is not in the set of allowed aromaticity models: "
                f"{ALLOWED_AROMATICITY_MODELS}"
            )

        self._aromaticity_model = aromaticity_model

    def _add_author(self, author: str):
        """
        Add an author to this force field. If this functional is called multiple times, all provided authors
        will be concatenated with the string " AND ". No redundancy checking is performed by this function.

        Parameters
        ----------
        author
            The author to add to this ForceField object
        """
        if self._author is None:
            self._author = author
        else:
            self._author += " AND " + author

    def _add_date(self, date: str):
        """
        Add an date to this force field. If this functional is called multiple times, all provided dates
        will be concatenated with the string " AND ". No redundancy checking is performed by this function.

        Parameters
        ----------
        date
            The author to add to this ForceField object
        """
        if self._date is None:
            self._date = date
        else:
            self._date += " AND " + date

    @property
    def author(self) -> str | None:
        """Returns the author data for this ForceField object. If not defined in any loaded files, this will be None.

        Returns
        -------
        author
            The author data for this force field.
        """
        return self._author

    @author.setter
    def author(self, author: str | None):
        """Set the author data for this ForceField object. If not defined in any loaded files, this will be None.

        Parameters
        ----------
        author
            The author data to set for this force field.
        """
        self._author = author

    @property
    def date(self) -> str | None:
        """Returns the date data for this ForceField object. If not defined in any loaded files, this will be None.

        Returns
        -------
        date
            The date data for this force field.
        """
        return self._date

    @date.setter
    def date(self, date: str | None):
        """Set the author data for this ForceField object. If not defined in any loaded files, this will be None.

        Parameters
        ----------
        date
            The date data to set for this force field.
        """
        self._date = date

    def _register_parameter_handler_classes(
        self,
        parameter_handler_classes: list[type[ParameterHandler]],
    ):
        """
        Register multiple ParameterHandler classes, ensuring they specify unique tags to process

        .. warning :: This API is experimental and subject to change.

        Parameters
        ----------
        parameter_handler_classes
           List of ParameterHandler classes to register for this ForceField.
        """
        for parameter_handler_class in parameter_handler_classes:
            tagname = parameter_handler_class._TAGNAME
            if tagname is not None:
                if tagname in self._parameter_handler_classes:
                    raise ParameterHandlerRegistrationError(
                        f"Attempting to register ParameterHandler {parameter_handler_class}, which provides a parser "
                        "for tag '{tagname}', but ParameterHandler {self._parameter_handler_classes[tagname]} has "
                        "already been registered to handle that tag."
                    )
                self._parameter_handler_classes[tagname] = parameter_handler_class

    def _register_parameter_io_handler_classes(
        self,
        parameter_io_handler_classes: list[type[ParameterIOHandler]],
    ):
        """
        Register multiple ParameterIOHandler classes, ensuring they specify unique suffixes

        .. warning :: This API is experimental and subject to change.

        Parameters
        ----------
        parameter_io_handler_classes
            All specified ParameterIOHandler classes will be registered as ways to translate to/from the object model
            to serialized parameter sets.

        Raises
        ------
        Exception
            If two ParameterIOHandlers are attempted to be registered for the same file format.

        """
        for parameter_io_handler_class in parameter_io_handler_classes:
            serialization_format = parameter_io_handler_class._FORMAT
            if serialization_format is not None:
                if serialization_format in self._parameter_io_handler_classes.keys():
                    raise ParameterHandlerRegistrationError(
                        f"Attempting to register ParameterIOHandler {parameter_io_handler_class}, which provides a IO "
                        "parser for format '{serialization_format}', but ParameterIOHandler "
                        "{self._parameter_io_handler_classes[serialization_format]} has already been registered to "
                        "handle that tag."
                    )
                self._parameter_io_handler_classes[serialization_format] = parameter_io_handler_class

    def register_parameter_handler(self, parameter_handler: ParameterHandler):
        """
        Register a new ParameterHandler for a specific tag, making it
        available for lookup in the ForceField.

        .. warning :: This API is experimental and subject to change.

        Parameters
        ----------
        parameter_handler
            The ParameterHandler to register. The TAGNAME attribute of this object will be used as the key for
            registration.

        """
        if not parameter_handler._TAGNAME:
            raise ValueError("Handler's `_TAGNAME` must not be `None`.")

        tagname = parameter_handler._TAGNAME
        if tagname in self._parameter_handlers.keys():
            raise ParameterHandlerRegistrationError(
                f"Tried to register parameter handler '{parameter_handler}' for tag '{tagname}', but "
                f"tag is already registered to {self._parameter_handlers[tagname]}"
            )

        self._parameter_handlers[parameter_handler._TAGNAME] = parameter_handler
        self._parameter_handler_classes[parameter_handler._TAGNAME] = type(parameter_handler)

    def register_parameter_io_handler(self, parameter_io_handler: ParameterIOHandler):
        """
        Register a new ParameterIOHandler, making it available for lookup in the ForceField.

        .. warning :: This API is experimental and subject to change.

        Parameters
        ----------
        parameter_io_handler
            The ParameterIOHandler to register. The FORMAT attribute of this object will be used
            to associate it to a file format/suffix.

        """
        io_format = parameter_io_handler._FORMAT
        if io_format in self._parameter_io_handlers.keys():
            raise ParameterHandlerRegistrationError(
                f"Tried to register parameter IO handler '{parameter_io_handler}' for tag '{io_format}', but "
                f"tag is already registered to {self._parameter_io_handlers[io_format]}"
            )
        self._parameter_io_handlers[io_format] = parameter_io_handler

    @property
    def registered_parameter_handlers(self) -> list[str]:
        """
        Return the list of registered parameter handlers by name

        .. warning :: This API is experimental and subject to change.

        Returns
        -------
            registered_parameter_handlers

        """
        return [*self._parameter_handlers.keys()]

    def get_parameter_handler(
        self,
        tagname: str,
        handler_kwargs: dict | None = None,
        allow_cosmetic_attributes: bool = False,
    ) -> ParameterHandler:
        """Retrieve the parameter handlers associated with the provided tagname.

        If the parameter handler has not yet been instantiated, it will be created and returned.
        If a parameter handler object already exists, it will be checked for compatibility
        and an Exception raised if it is incompatible with the provided kwargs. If compatible, the
        existing ParameterHandler will be returned.

        Parameters
        ----------
        tagname
            The name of the parameter to be handled.
        handler_kwargs
            Dict to be passed to the handler for construction or checking compatibility. If this is None and no
            existing ParameterHandler exists for the desired tag, a handler will be initialized with all default
            values. If this is None and a handler for the desired tag exists, the existing ParameterHandler will
            be returned.
        allow_cosmetic_attributes
            Whether to permit non-spec kwargs in smirnoff_data.

        Returns
        -------
        handler

        Raises
        ------
        KeyError
            If there is no ParameterHandler for the given tagname
        """
        # If there are no kwargs for the handler, initialize handler_kwargs as an empty dict
        skip_version_check = False
        if handler_kwargs is None:
            handler_kwargs = dict()
            skip_version_check = True

        # Ensure that the ForceField has a ParameterHandler class registered that can handle this tag
        ph_class = self._get_parameter_handler_class(tagname)

        if tagname in self._parameter_handlers:
            # If a handler of this class already exists, ensure that the two handlers encode compatible science
            old_handler = self._parameter_handlers[tagname]
            # If no handler kwargs were provided, skip the compatibility check
            if handler_kwargs != {}:
                # Initialize a new instance of this parameter handler class with the given kwargs
                new_handler = ph_class(
                    **handler_kwargs,
                    allow_cosmetic_attributes=allow_cosmetic_attributes,
                    skip_version_check=skip_version_check,
                )
                old_handler.check_handler_compatibility(new_handler)
            return_handler = old_handler
        elif tagname in self._parameter_handler_classes:
            # Otherwise, register this handler in the force field
            # Initialize a new instance of this parameter handler class with the given kwargs
            new_handler = ph_class(
                **handler_kwargs,
                allow_cosmetic_attributes=allow_cosmetic_attributes,
                skip_version_check=skip_version_check,
            )
            self.register_parameter_handler(new_handler)
            return_handler = new_handler

        return return_handler

    def get_parameter_io_handler(self, io_format: str) -> ParameterIOHandler:
        """Retrieve the parameter handlers associated with the provided tagname.
        If the parameter IO handler has not yet been instantiated, it will be created.

        Parameters
        ----------
        io_format
            The name of the io format to be handled.

        Returns
        -------
        io_handler

        Raises
        ------
        KeyError
            If there is no ParameterIOHandler for the given tagname
        """
        # Uppercase the format string to avoid case mismatches
        io_format = io_format.upper()

        # Remove "." (ParameterIOHandler tags do not include it)
        io_format = io_format.strip(".")

        # Find or initialize ParameterIOHandler for this format
        if io_format in self._parameter_io_handlers.keys():
            return self._parameter_io_handlers[io_format]
        elif io_format in self._parameter_io_handler_classes.keys():
            new_handler_class = self._parameter_io_handler_classes[io_format]
            io_handler = new_handler_class()
            self.register_parameter_io_handler(io_handler)
            return io_handler
        else:
            raise KeyError(
                f"Cannot find a registered parameter IO handler for format '{io_format}'\n"
                f"Registered parameter IO handlers: {self._parameter_io_handlers.keys()}\n"
            )

    def deregister_parameter_handler(
        self,
        handler: str | ParameterHandler,
    ):
        """
        Deregister a parameter handler specified by tag name, class, or instance.

        Parameters
        ----------
        handler
            The handler to deregister.
        """
        if isinstance(handler, ParameterHandler):
            tagname = handler.TAGNAME
        elif isinstance(handler, str):  # Catch case of name (as str) before checking subclass
            tagname = handler
        elif issubclass(handler, ParameterHandler):
            tagname = handler._TAGNAME
        else:
            tagname = handler
        del self._parameter_handlers[tagname]

    def parse_sources(self, sources, allow_cosmetic_attributes: bool = True):
        """Parse a SMIRNOFF force field definition.

        Parameters
        ----------
        sources
            An iterable of files defining the SMIRNOFF force field to be loaded.
            Currently, only `the SMIRNOFF XML format <https://openforcefield.github.io/standards/standards/smirnoff/>`_
            is supported.  Each entry may be an absolute file path, a path relative to the current working directory, a
            path relative to this module's data subdirectory (for built in force fields), or an open file-like object
            with a ``read()`` method from which the force field XML data can be loaded.  If multiple files are
            specified, any top-level tags that are repeated will be merged if they are compatible, with files appearing
            later in the sequence resulting in parameters that have higher precedence.  Support for multiple files is
            primarily intended to allow solvent parameters to be specified by listing them last in the sequence.
        allow_cosmetic_attributes
            Whether to permit non-spec kwargs present in the source.

        Notes
        -----

            * New SMIRNOFF sections are handled independently, as if they were specified in the same file.
            * If a SMIRNOFF section that has already been read appears again, its definitions are appended to the end
                of the previously-read definitions if the sections are configured with compatible attributes;
                otherwise, an ``IncompatibleTagException`` is raised.

        """
        # Ensure that we are working with an iterable
        try:
            sources = iter(sources)
        except TypeError:
            # Make iterable object
            sources = [sources]

        # TODO: If a non-first source fails here, the force field might be partially modified
        for source in sources:
            smirnoff_data = self.parse_smirnoff_from_source(source)
            self._load_smirnoff_data(smirnoff_data, allow_cosmetic_attributes=allow_cosmetic_attributes)

    def _to_smirnoff_data(self, discard_cosmetic_attributes: bool = False) -> dict:
        """
        Convert this ForceField and all related ParameterHandlers to a dict representing a SMIRNOFF
        data object.

        Parameters
        ----------
        discard_cosmetic_attributes
            Whether to discard any non-spec attributes stored in the ForceField.

        Returns
        -------
        smirnoff_data
            A nested dict representing this ForceField as a SMIRNOFF data object.

        """
        l1_dict: dict[str, str | dict] = dict()

        # Assume we will write out SMIRNOFF data in compliance with the max supported spec version
        l1_dict["version"] = str(self._MAX_SUPPORTED_SMIRNOFF_VERSION)

        # Write out the aromaticity model used
        l1_dict["aromaticity_model"] = self._aromaticity_model

        # Write out author and date (if they have been set)
        if self._author is not None:
            l1_dict["Author"] = self._author

        # Write out author and date (if they have been set)
        if self._date is not None:
            l1_dict["Date"] = self._date

        for parameter_handler in self._parameter_handlers.values():
            # If _TAGNAME is None, the default value, an error should have been
            # thrown upon registering it, so assume it is str here
            handler_tag: str = parameter_handler._TAGNAME  # type: ignore[assignment]
            l1_dict[handler_tag] = parameter_handler.to_dict(discard_cosmetic_attributes=discard_cosmetic_attributes)

        smirnoff_data: dict[str, dict] = dict()
        smirnoff_data["SMIRNOFF"] = l1_dict

        return convert_all_quantities_to_string(smirnoff_data)

    # TODO: Should we call this "from_dict"?
    def _load_smirnoff_data(
        self,
        smirnoff_data: dict,
        allow_cosmetic_attributes: bool = False,
    ):
        """
        Add parameters from a SMIRNOFF-format data structure to this ForceField.

        Parameters
        ----------
        smirnoff_data
            A representation of a SMIRNOFF-format data structure. Begins at top-level 'SMIRNOFF' key.
        allow_cosmetic_attributes
            Whether to permit non-spec kwargs in smirnoff_data.
        """
        import packaging.version

        # Check that the SMIRNOFF version of this data structure is supported by this ForceField implementation

        if "SMIRNOFF" in smirnoff_data:
            version = smirnoff_data["SMIRNOFF"]["version"]
        elif "SMIRFF" in smirnoff_data:
            version = smirnoff_data["SMIRFF"]["version"]
        else:
            raise SMIRNOFFParseError("'version' attribute must be specified in SMIRNOFF tag")

        self._check_smirnoff_version_compatibility(str(version))

        # Convert 0.1 spec files to 0.3 SMIRNOFF data format by converting
        # from 0.1 spec to 0.2, then 0.2 to 0.3
        if packaging.version.parse(str(version)) == packaging.version.parse("0.1"):
            # NOTE: This will convert the top-level "SMIRFF" tag to "SMIRNOFF"
            smirnoff_data = convert_0_1_smirnoff_to_0_2(smirnoff_data)
            smirnoff_data = convert_0_2_smirnoff_to_0_3(smirnoff_data)

        # Convert 0.2 spec files to 0.3 SMIRNOFF data format by removing units
        # from section headers and adding them to quantity strings at all levels.
        elif packaging.version.parse(str(version)) == packaging.version.parse("0.2"):
            smirnoff_data = convert_0_2_smirnoff_to_0_3(smirnoff_data)

        # Ensure that SMIRNOFF is a top-level key of the dict
        if "SMIRNOFF" not in smirnoff_data:
            raise SMIRNOFFParseError("'SMIRNOFF' must be a top-level key in the SMIRNOFF object model")

        # Check that the aromaticity model required by this parameter set is compatible with
        # others loaded by this ForceField
        if "aromaticity_model" in smirnoff_data["SMIRNOFF"]:
            aromaticity_model = smirnoff_data["SMIRNOFF"]["aromaticity_model"]
            self.aromaticity_model = aromaticity_model

        elif self._aromaticity_model is None:
            raise SMIRNOFFParseError(
                "'aromaticity_model' attribute must be specified in SMIRNOFF "
                "tag, or contained in a previously-loaded SMIRNOFF data source"
            )

        if "Author" in smirnoff_data["SMIRNOFF"]:
            self._add_author(smirnoff_data["SMIRNOFF"]["Author"])

        if "Date" in smirnoff_data["SMIRNOFF"]:
            self._add_date(smirnoff_data["SMIRNOFF"]["Date"])

        # Go through the whole SMIRNOFF data structure, trying to convert all strings to Quantity
        smirnoff_data = convert_all_strings_to_quantity(
            smirnoff_data,
            ignore_keys=["smirks", "name", "id", "parent_id"],
        )

        # Go through the subsections, delegating each to the proper ParameterHandler

        # Define keys which are expected from the spec, but are not parameter sections
        l1_spec_keys = ["Author", "Date", "version", "aromaticity_model"]
        # TODO: Throw SMIRNOFFSpecError for unrecognized keywords

        for parameter_name in smirnoff_data["SMIRNOFF"]:
            # Skip (for now) cosmetic l1 items. They're handled above
            if parameter_name in l1_spec_keys:
                continue
            # Handle cases where a parameter name has no info (eg. ToolkitAM1BCC)
            if smirnoff_data["SMIRNOFF"][parameter_name] is None:
                # "get"ting the parameter handler here will also initialize it.
                _ = self.get_parameter_handler(parameter_name, {})
                continue

            # Otherwise, we expect this l1_key to correspond to a ParameterHandler
            section_dict = smirnoff_data["SMIRNOFF"][parameter_name]

            # TODO: Implement a ParameterHandler.from_dict() that knows how to deserialize itself for extensibility.
            #       We could let it load the ParameterTypes from the dict and append them to the existing handler
            #       after verifying that they are compatible.

            # Get the parameter types serialization that is not passed to the ParameterHandler constructor.
            ph_class = self._get_parameter_handler_class(parameter_name)
            try:
                infotype = ph_class._INFOTYPE
                parameter_list_tagname = infotype._ELEMENT_NAME  # type: ignore[union-attr]
            except AttributeError:
                # The ParameterHandler doesn't have ParameterTypes (e.g. ToolkitAM1BCCHandler).
                parameter_list_dict = {}
            else:
                parameter_list_dict = section_dict.pop(parameter_list_tagname, {})

            # Must be wrapped into its own tag.
            # Assumes that parameter_list_dict is always a list

            # If the parameter list isn't empty, it must be transferred into its own tag.
            # This is necessary for deserializing SMIRNOFF force field sections which may or may
            # not have any smirks-based elements (like an empty ChargeIncrementModel section)
            if parameter_list_dict != {}:
                parameter_list_dict = {parameter_list_tagname: parameter_list_dict}

            # Retrieve or create parameter handler, passing in section_dict to check for
            # compatibility if a handler for this parameter name already exists
            handler = self.get_parameter_handler(
                parameter_name,
                section_dict,
                allow_cosmetic_attributes=allow_cosmetic_attributes,
            )
            handler._add_parameters(parameter_list_dict, allow_cosmetic_attributes=allow_cosmetic_attributes)

    def parse_smirnoff_from_source(self, source: str | IO) -> dict:
        """
        Reads a SMIRNOFF data structure from a source, which can be one of many types.

        Parameters
        ----------
        source
            File defining the SMIRNOFF force field to be loaded Currently, only `the SMIRNOFF XML format
            <https://openforcefield.github.io/standards/standards/smirnoff/>`_ is supported.  The file may be an
            absolute file path, a path relative to the current working directory, a path relative to this module's data
            subdirectory (for built in force fields), or an open file-like object with a ``read()`` method from which
            the force field XML data can be loaded.

        Returns
        -------
        smirnoff_data
            A representation of a SMIRNOFF-format data structure. Begins at top-level 'SMIRNOFF' key.

        """
        # First, see if a file exists with a name `source` in the current directory or in directories known to the
        # plugin system. It could also be a raw XML-like string ...
        if isinstance(source, str):
            # Try first the simple path.
            searched_dirs_paths: list[str] = [os.getcwd()]
            # Then try a relative file path w.r.t. an installed directory.
            searched_dirs_paths.extend(_get_installed_offxml_dir_paths())

            # Determine the actual path of the file.
            # TODO: What is desired toolkit behavior if two files with the desired name are available?
            for dir_path in searched_dirs_paths:
                for file_path in pathlib.Path(dir_path).glob("*.offxml"):
                    if str(file_path).lower().endswith(source.lower()):
                        source = str(file_path.absolute())
                        break

        else:
            # ... or a file-like object, in which case we shouldn't look through the plugin system.
            # if it's raw bytes, no need to search for paths, either.
            searched_dirs_paths = list()

        # Process all SMIRNOFF definition files or objects
        # QUESTION: Allow users to specify force field URLs so they can pull force field definitions from the web too?
        io_formats_to_try = self._parameter_io_handler_classes.keys()

        # Parse content depending on type
        for parameter_io_format in io_formats_to_try:
            parameter_io_handler = self.get_parameter_io_handler(parameter_io_format)

            # Try parsing as a force field file or file-like object
            try:
                smirnoff_data = parameter_io_handler.parse_file(source)
                return smirnoff_data
            except SMIRNOFFParseError as e:
                exception_type = type(e)
                exception_context = "while trying to parse source as an object"
                exception_msg = e.msg
            except OSError:
                # If this is not a file path or a file handle, attempt parsing as a string.
                # TODO: Do we actually support parsing bytes?
                try:
                    smirnoff_data = parameter_io_handler.parse_string(source)
                    return smirnoff_data
                except SMIRNOFFParseError as e:
                    exception_type = type(e)
                    exception_context = "while trying to parse source as a file"
                    exception_msg = e.args[0]

        # If we haven't returned by now, the parsing was unsuccessful
        # There is different parsing behavior for str and file-like objects, so raise errors separately
        if isinstance(source, str):
            pretty_searched_paths = "\n    ".join(searched_dirs_paths)
            msg = (
                f"Source '{source}' could not be read. If this is a file, ensure that the path is correct.\n"
                f"Looked in the following paths and found no files named '{source}':"
                f"\n    {pretty_searched_paths}\n"
                f"If '{source}' is present as a file, ensure it is in a known SMIRNOFF encoding.\n"
                f"Valid formats are: {[*self._parameter_io_handler_classes.keys()]}\n"
                f"Parsing failed {exception_context} with the following exception and "
                f"message:\n{exception_type}\n{exception_msg}\n"
            )

        else:
            msg = (
                f"Source '{source}' could not be read.\n"
                f"Parsing failed {exception_context} with the following exception and message:\n"
                f"{exception_type}\n{exception_msg}\n"
            )

        raise OSError(msg)

    def to_string(
        self,
        io_format: str | ParameterIOHandler = "XML",
        discard_cosmetic_attributes: bool = False,
    ) -> str:
        """
        Write this Forcefield and all its associated parameters to a string in a given format which
        complies with the SMIRNOFF spec.


        Parameters
        ----------
        io_format
            The serialization format to write to
        discard_cosmetic_attributes
            Whether to discard any non-spec attributes stored in the ForceField.

        Returns
        -------
        forcefield_string
            The string representation of the serialized force field
        """
        # Resolve which IO handler to use
        if isinstance(io_format, ParameterIOHandler):
            io_handler = io_format
        else:
            io_handler = self.get_parameter_io_handler(io_format)

        smirnoff_data = self._to_smirnoff_data(discard_cosmetic_attributes=discard_cosmetic_attributes)
        string_data = io_handler.to_string(smirnoff_data)
        return string_data

    def to_file(
        self,
        filename: str,
        io_format: str | ParameterIOHandler | None = None,
        discard_cosmetic_attributes: bool = False,
    ) -> None:
        """
        Write this Forcefield and all its associated parameters to a string in a given format which
        complies with the SMIRNOFF spec.

        Parameters
        ----------
        filename
            The filename to write to
        io_format
            The serialization format to write out. If None, will attempt to be inferred from the filename.
        discard_cosmetic_attributes
            Whether to discard any non-spec attributes stored in the ForceField.

        """
        if io_format is None:
            _, io_format = os.path.splitext(filename)

        # Resolve which IO handler to use
        if isinstance(io_format, ParameterIOHandler):
            io_handler = io_format
        else:
            # Handle the fact that .offxml is the same as .xml
            if io_format.lower() == "offxml" or io_format.lower() == ".offxml":
                io_format = "xml"
            io_handler = self.get_parameter_io_handler(io_format)

        # Write out the SMIRNOFF data to the IOHandler
        smirnoff_data = self._to_smirnoff_data(discard_cosmetic_attributes=discard_cosmetic_attributes)
        io_handler.to_file(filename, smirnoff_data)

    def combine(
        self,
        other: "ForceField",
    ) -> "ForceField":
        """
        Combine this `ForceField` with another `ForceField`, returning a new `ForceField`.

        The same rules as `ForceField.__init__` are followed.
        """
        import copy

        combined = copy.deepcopy(self)

        combined._load_smirnoff_data(
            smirnoff_data=other._to_smirnoff_data(
                discard_cosmetic_attributes=False,
            ),
            allow_cosmetic_attributes=True,
        )

        return combined

    # TODO: Should we also accept a Molecule as an alternative to a Topology?
    @requires_package("openmm")
    def create_openmm_system(
        self,
        topology: "Topology",
        *,
        toolkit_registry: Union["ToolkitRegistry", "ToolkitWrapper"] | None = None,
        charge_from_molecules: list["Molecule"] | None = None,
        partial_bond_orders_from_molecules: list["Molecule"] | None = None,
        allow_nonintegral_charges: bool = False,
    ) -> "openmm.System":
        """Create an OpenMM System from this ForceField and a Topology.

        Note that most force fields specify their own partial charges, and any
        partial charges defined on the ``Molecule`` objects in the topology are
        ignored. To use custom partial charges, see the
        ``charge_from_molecules`` argument.

        Parameters
        ----------
        topology
            The ``Topology`` which is to be parameterized with this
            ``ForceField``.
        toolkit_registry
            The toolkit registry to use for parametrization (eg, for calculating
            partial charges and partial bond orders)
        charge_from_molecules
            Take partial charges from the input topology rather than calculating
            them. This may be useful for avoiding recalculating charges, but
            take care to ensure that your charges are appropriate for the force
            field.
        partial_bond_orders_from_molecules
            Take partial bond orders from the input topology rather than
            calculating them. This may be useful for avoiding recalculating
            PBOs, but take to ensure that they are appropriate for the
            force field.
        allow_nonintegral_charges
            Allow charges that do not sum to an integer.
        """
        return self.create_interchange(
            topology,
            toolkit_registry,
            charge_from_molecules=charge_from_molecules,
            partial_bond_orders_from_molecules=partial_bond_orders_from_molecules,
            allow_nonintegral_charges=allow_nonintegral_charges,
        ).to_openmm(combine_nonbonded_forces=True)

    @requires_package("openff.interchange")
    def create_interchange(
        self,
        topology: "Topology",
        toolkit_registry: Union["ToolkitRegistry", "ToolkitWrapper"] | None = None,
        charge_from_molecules: list["Molecule"] | None = None,
        partial_bond_orders_from_molecules: list["Molecule"] | None = None,
        allow_nonintegral_charges: bool = False,
    ) -> "Interchange":
        """
        Create an Interchange object from a ForceField, Topology, and (optionally) box vectors.

        WARNING: This API and functionality are experimental and not suitable for production.

        Parameters
        ----------
        topology
            The topology to create this `Interchange` object from.
        toolkit_registry
            The toolkit registry to use for parametrization (eg, for calculating
            partial charges and partial bond orders)
        charge_from_molecules
            Take charges from the input topology rather than calculating them.
            This may be useful for avoiding recalculating charges, but take care
            to ensure that your charges are appropriate for the force field.
        partial_bond_orders_from_molecules
            Take partial bond orders from the input topology rather than
            calculating them. This may be useful for avoiding recalculating
            PBOs, but take to ensure that they are appropriate for the
            force field.
        allow_nonintegral_charges
            Allow charges that do not sum to an integer.

        Returns
        -------
        interchange
            An `Interchange` object resulting from applying this `ForceField` to a `Topology`.

        """
        from openff.interchange import Interchange

        from openff.toolkit.utils.toolkit_registry import toolkit_registry_manager

        if toolkit_registry is not None:
            used_registry = toolkit_registry
        else:
            from openff.toolkit.utils.toolkits import GLOBAL_TOOLKIT_REGISTRY

            used_registry = GLOBAL_TOOLKIT_REGISTRY

        with toolkit_registry_manager(used_registry):
            return Interchange.from_smirnoff(
                force_field=self,
                topology=topology,
                charge_from_molecules=charge_from_molecules,
                partial_bond_orders_from_molecules=partial_bond_orders_from_molecules,
                allow_nonintegral_charges=allow_nonintegral_charges,
            )

    def label_molecules(self, topology: "Topology") -> list[dict[str, "ValenceDict"]]:
        """
        Return labels for a list of molecules corresponding to parameters from this force field.

        For each molecule, a dictionary of force types is returned, and for each force type,
        each force term is provided with the atoms involved, the parameter id assigned, and the corresponding SMIRKS.

        .. todo::
           What is the most useful API for this method?

           - Should we instead accept :class:`Molecule` objects as input and
             individually return labels?
           - Should we attach the labels to the :class:`Molecule` object?
           - Or should we label all interactions in a :class:`Topology` instead
             of just labeling its ``unique_molecules``?

        Parameters
        ----------
        topology
            A Topology object containing one or more unique molecules to be labeled

        Returns
        -------
        molecule_labels
            List of labels for unique molecules. Each entry in the list
            corresponds to one unique molecule in the Topology and is a
            dictionary keyed by force type, i.e., ``molecule_labels[0]
            ['HarmonicBondForce']`` gives details for the harmonic bond
            parameters for the first molecule. Each element is a list of the
            form: ``[ ( [ atom1, ..., atomN], parameter_id, SMIRKS), ... ]``.
        """
        from openff.toolkit.typing.engines.smirnoff.parameters import VirtualSiteHandler

        # Loop over molecules and label
        molecule_labels = list()

        # TODO: This was previously ... enumerate(topology.reference_molecules). It's currently
        # unclear if this should be topology.unique_molecules instead, since that might be faster
        # (if also modifying this to label _all_ duplicates of each unique molecule)
        for molecule in topology.molecules:
            current_molecule_labels = dict()
            for tag, parameter_handler in self._parameter_handlers.items():
                param_is_list = False

                if type(parameter_handler) is VirtualSiteHandler:
                    param_is_list = True

                matches = parameter_handler.find_matches(molecule.to_topology())

                # Remove the chemical environment matches from the
                # matched results.

                # Because we sometimes need to enforce atom ordering,
                # `matches` here is not a normal `dict`, but rather
                # one that transforms keys in arbitrary ways. Thus,
                # we need to make a copy of its specific class here.

                parameter_matches = matches.__class__()

                # Now make parameter_matches into a dict mapping
                # match objects to ParameterTypes

                if param_is_list:
                    for match in matches:
                        parameter_matches[match] = [m.parameter_type for m in matches[match]]
                else:
                    for match in matches:
                        parameter_matches[match] = matches[match].parameter_type

                current_molecule_labels[tag] = parameter_matches

            molecule_labels.append(current_molecule_labels)
        return molecule_labels

    def _get_parameter_handler_class(self, tagname: str) -> type[ParameterHandler]:
        """Retrieve the ParameterHandler class associated to the tagname and throw a custom error if not found."""
        try:
            ph_class = self._parameter_handler_classes[tagname]
        except KeyError:
            msg = f"Cannot find a registered parameter handler class for tag '{tagname}'\n"
            msg += f"Known parameter handler class tags are {self._parameter_handler_classes.keys()}"
            raise KeyError(msg)
        return ph_class

    @requires_package("openff.interchange")
    def get_partial_charges(self, molecule: "Molecule", **kwargs: Any) -> Quantity:
        """Generate the partial charges for the given molecule in this force field.

        Parameters
        ----------
        molecule
            The ``Molecule`` corresponding to the system to be parameterized
        toolkit_registry
            The toolkit registry to use for operations like conformer generation and
            partial charge assignment.

        Returns
        -------
        charges
            The partial charges of the provided molecule in this force field.

        Raises
        ------
        PartialChargeVirtualSitesError
            If the ``ForceField`` applies virtual sites to the ``Molecule``.
            ``get_partial_charges`` cannot identify which virtual site charges
            may belong to which atoms in this case.

        Other exceptions
            As any ``ParameterHandler`` may in principle modify charges, the entire
            force field must be applied to the molecule to produce the charges.
            Calls to this method from incorrectly or incompletely specified ``ForceField``
            objects thus may raise an exception.

        Examples
        --------

        >>> from openff.toolkit import ForceField, Molecule
        >>> ethanol = Molecule.from_smiles('CCO')
        >>> force_field = ForceField('openff-2.0.0.offxml')

        Assign partial charges to the molecule according to the force field:

        >>> ethanol.partial_charges = force_field.get_partial_charges(ethanol)

        Use the assigned partial charges when creating an OpenMM ``System``:

        >>> topology = ethanol.to_topology()
        >>> system = force_field.create_openmm_system(
        ...    topology,
        ...    charge_from_molecules=[ethanol]
        ... )

        This is especially useful when you want to create multiple systems
        with the same molecule or molecules, as it allows the expensive
        charge calculation to be cached.

        """
        from openff.toolkit import Molecule, unit

        if not isinstance(molecule, Molecule):
            raise ValueError(
                f"`molecule` argument must be a `Molecule` or subclass object. Found type {type(molecule)}"
            )

        return Quantity(
            [
                c.m
                for c in self.create_interchange(
                    topology=molecule.to_topology(),
                    **kwargs,
                )["Electrostatics"].charges.values()
            ],
            unit.elementary_charge,
        )

    def __getitem__(self, val: str | ParameterHandler) -> ParameterHandler:
        """
        Syntax sugar for lookikng up a ParameterHandler. Note that only
        string-based lookups are currently supported.
        """
        if isinstance(val, str):
            if val in self._parameter_handlers:
                return self.get_parameter_handler(val)
            else:
                raise KeyError(f"Parameter handler with name '{val}' not found.")
        elif isinstance(val, ParameterHandler) or issubclass(val, ParameterHandler):
            raise NotImplementedError

    def __hash__(self) -> int:
        """Deterministically hash a ForceField object

        Notable behavior:
          * `author` and `date` are stripped from the ForceField
          * `id` and `parent_id` are stripped from each ParameterType"""
        import copy

        # Completely re-constructing the force field may be overkill
        # compared to deepcopying and modifying, but is not currently slow
        ff_copy = ForceField()
        ff_copy.date = ""
        ff_copy.author = ""

        param_attrs_to_strip = ["_id", "_parent_id"]

        for handler_name in self.registered_parameter_handlers:
            handler = copy.deepcopy(self.get_parameter_handler(handler_name))

            for param in handler._parameters:
                for attr in param_attrs_to_strip:
                    # param.__dict__.pop(attr, None) may be faster
                    # https://stackoverflow.com/a/42303681/4248961
                    if hasattr(param, attr):
                        delattr(param, attr)

            ff_copy.register_parameter_handler(handler)

        return hash(ff_copy.to_string(discard_cosmetic_attributes=True))



#!/usr/bin/env python3
"""Copy the openff-toolkit examples suite to a local directory"""

import argparse
from os import environ
from pathlib import Path
from shutil import copytree, ignore_patterns


class NoExamplesLibraryError(Exception):
    """Slightly more descriptive error for when the examples library is not found"""

    def __init__(self):
        message = "\n".join(
            (
                "Could not find examples library.",
                "",
                "Is the openff-toolkit-examples Conda package installed?",
            )
        )
        super().__init__(message)


def main():
    """Entrypoint for the script"""
    parser = argparse.ArgumentParser(description=__doc__)

    default_target = "./examples/"
    parser.add_argument(
        "--target",
        default=default_target,
        type=Path,
        help=f"Target directory to store the examples suite. Defaults to {default_target}",
    )

    parser.add_argument(
        "--include-deprecated",
        action="store_true",
        help="Copy over deprecated examples in addition to current examples",
    )

    args = parser.parse_args()

    if args.include_deprecated:
        copytree_ignore = None
    else:
        copytree_ignore = ignore_patterns("deprecated")

    target_path = args.target
    library_path = Path(environ["CONDA_PREFIX"]) / "share/openff-toolkit/examples"

    try:
        copytree(library_path, target_path, ignore=copytree_ignore)
    except FileNotFoundError as err:
        if Path(err.filename) == library_path and err.filename2 is None:
            raise NoExamplesLibraryError() from err
        raise err


if __name__ == "__main__":
    main()



import json

import numpy as np
import openmm
import pytest
from openff.units.openmm import to_openmm
from openmm import unit as openmm_unit

from openff.toolkit._tests.utils import get_data_file_path, requires_rdkit
from openff.toolkit.topology import Molecule, Topology
from openff.toolkit.typing.engines.smirnoff import ForceField


@requires_rdkit
@pytest.mark.parametrize("constrained", [True, False])
@pytest.mark.parametrize(
    "mol",
    [
        "ethanol.sdf",
        "methane_multiconformer.sdf",
        "CID20742535_anion.sdf",
    ],
)
def test_reference(constrained, mol):
    """Minimal regression test comparing molecule energies to energies computed
    by version 0.8.0 of the toolkit"""
    # TODO: Also test periodic vs. vacuum
    with open(get_data_file_path("reference_energies/reference_after_1007.json")) as fi:
        reference = json.loads(fi.read())

    name = mol + "_"
    if not constrained:
        name += "un"
    name += "constrained"
    reference_energy = reference[name]

    omm_sys, positions, off_top = _build_system(
        mol=mol,
        constrained=constrained,
    )

    simulation = _build_simulation(omm_sys=omm_sys, off_top=off_top)
    derived_energy = _get_energy(simulation=simulation, positions=positions)

    np.testing.assert_almost_equal(
        actual=derived_energy.value_in_unit(openmm_unit.kilojoule_per_mole),
        desired=reference_energy,
        decimal=5,
    )


def generate_reference():
    """Function to generate reference files, if this script is called directly by a Python interpreter.
    This is ignored by pytest while running tests."""
    reference = dict()
    for mol in [
        "ethanol.sdf",
        "methane_multiconformer.sdf",
        "CID20742535_anion.sdf",
    ]:
        for constrained in [True, False]:
            omm_sys, positions, off_top = _build_system(
                mol=mol,
                constrained=constrained,
            )

            simulation = _build_simulation(omm_sys=omm_sys, off_top=off_top)
            energy = _get_energy(simulation=simulation, positions=positions)

            name = mol + "_"
            if not constrained:
                name += "un"
            name += "constrained"
            reference.update({name: energy.value_in_unit(openmm_unit.kilojoule_per_mole)})

    import openff.toolkit

    toolkit_version = openff.toolkit.__version__

    with open(f"reference_{toolkit_version}.json", "w") as json_out:
        json.dump(reference, json_out)


def _build_system(mol, constrained):
    if constrained:
        parsley = ForceField("openff-1.0.0.offxml")
    else:
        parsley = ForceField("openff_unconstrained-1.0.0.offxml")

    mol = Molecule.from_file(get_data_file_path("molecules/" + mol), file_format="sdf")

    if type(mol) is Molecule:
        off_top = mol.to_topology()
        positions = mol.conformers[0]
    elif type(mol) is list:
        # methane_multiconformer case is a list of two mols
        off_top = Topology()
        for mol_i in mol:
            off_top.add_molecule(mol_i)
        positions = np.vstack([mol[0].conformers[0], mol[1].conformers[0]])

    from openff.toolkit.utils.toolkits import (
        AmberToolsToolkitWrapper,
        RDKitToolkitWrapper,
        ToolkitRegistry,
    )

    toolkit_registry = ToolkitRegistry(toolkit_precedence=[RDKitToolkitWrapper, AmberToolsToolkitWrapper])

    omm_sys = parsley.create_openmm_system(off_top, toolkit_registry=toolkit_registry)

    return omm_sys, positions, off_top


def _build_simulation(omm_sys, off_top):
    """Given an OpenMM System, initialize a barebones OpenMM Simulation."""
    # Use OpenMM to compute initial and minimized energy for all conformers
    integrator = openmm.VerletIntegrator(1 * openmm_unit.femtoseconds)
    platform = openmm.Platform.getPlatformByName("Reference")
    omm_top = off_top.to_openmm()
    simulation = openmm.app.Simulation(omm_top, omm_sys, integrator, platform)

    return simulation


def _get_energy(simulation, positions):
    """Given an OpenMM simulation and position, return its energy"""
    simulation.context.setPositions(to_openmm(positions))
    state = simulation.context.getState(getEnergy=True, getPositions=True)
    energy = state.getPotentialEnergy()

    return energy


if __name__ == "__main__":
    generate_reference()



#!/bin/env python

import numpy as np
from openff.toolkit.typing.engines.smirnoff.forcefield import ForceField
from openff.toolkit.typing.chemistry import environment

# Function definitions for parsing sections within parameter file
def _parse_nonbon_line( line ):
    """Parse an AMBER frcmod nonbon line and return relevant parameters in a dictionary. AMBER uses rmin_half and epsilon in angstroms and kilocalories per mole."""
    tmp = line.split()
    params = {}
    params['smirks'] = tmp[0]
    params['rmin_half'] = tmp[1]
    params['epsilon'] = tmp[2]

    return params

def _parse_bond_line( line ):
    """Parse an AMBER frcmod BOND line and return relevant parameters in a dictionary. AMBER uses length and force constant, with the factor of two dropped. Here we multiply by the factor of two before returning. Units are angstroms and kilocalories per mole per square angstrom."""

    tmp = line.split()
    params = {}
    params['smirks'] = tmp[0]
    params['k'] = str(2*float(tmp[1]))
    params['length'] = tmp[2]
    return params

def _parse_angl_line( line ):
    """Parse an AMBER frcmod ANGL line and return relevant parameters in a dictionary. AMBER uses angle and force constant, with the factor of two dropped. Here we multiply by the factor of two before returning. Units are degrees and kilocalories per mole."""

    tmp = line.split()
    params = {}
    params['smirks'] = tmp[0]
    params['k'] = str(2*float(tmp[1]))
    params['angle'] = tmp[2]
    return params

def _parse_dihe_line( line ):
    """Parse an AMBER frcmod DIHE line and return relevant parameters in a dictionary. Units for k are kilocalories per mole."""

    tmp = line.split()
    params = {}
    params['smirks'] = tmp[0]
    params['idivf1'] = tmp[1]
    params['k1'] = tmp[2]
    params['phase1'] = tmp[3]
    params['periodicity1'] = str(int(np.abs(float(tmp[4]))))
    return params

def _parse_impr_line( line ):
    """Parse an AMBER frcmod DIHE line and return relevant parameters in a dictionary. Units for k are kilocalories per mole."""

    tmp = line.split()
    params = {}
    params['smirks'] = tmp[0]
    params['k1'] = tmp[1]
    params['phase1'] = tmp[2]
    params['periodicity1'] = str(int(np.abs(float(tmp[3]))))
    return params

def add_date_and_author(inxml, date, author):
    """
    Updates the template xml file with the date and authors in the
    input Frcmodish file.
    Parameters
    ----------
    inxml: str, template xml file
    date: str, date from input Frcmod file
    author: str, author list from input Frcmod file
    """
    # read input file
    f = open(inxml,'r')
    input_lines = f.readlines()
    f.close()

    output_lines = list()
    # Save lines, only change those for Date and Author
    for l in input_lines:
        start = l.strip().split('>')[0]
        if start == '<Date':
            output_lines.append(f"<Date>{date.strip()}</Date>\n")
        elif start == '<Author':
            output_lines.append(f"<Author>{author.strip()}</Author>\n")
        else:
            output_lines.append(l)

    # write fixed lines to ffxml tempate
    f = open(inxml,'w')
    f.writelines(output_lines)
    f.close()


# Main conversion functionality
def convert_frcmod_to_ffxml( infile, inxml, outxml ):
    """Convert a modified AMBER frcmod (with SMIRKS replacing atom types) to SMIRNOFF ffxml format by inserting parameters into a template ffxml file.

    Parameters
    ----------
    infile : str
        File name of input SMIRKS-ified frcmod file containing parameters
    inxml : str
        File name of template SMIRNOFF FFXML file into which to insert these parameters.
    outxml : str
        File name of resulting output SMIRNOFF FFXML

    Notes:
    -------
    Input XML file will normally be the template of a SMIRNOFF XML file without any parameters present (but with requisite force types already specified).
    """

    # Obtain sections from target file
    file = open(infile, 'r')
    text = file.readlines()
    file.close()
    sections = {}
    # Section names from frcmod which we will parse
    secnames = ['NONBON', 'BOND', 'ANGL', 'IMPR', 'DIHE']
    # Tags that will be used in the FFXML for these (same order)
    tag = ['Atom', 'Bond', 'Angle', 'Improper', 'Proper']
    # Force names in the FFXML (same order)
    force_section = ['NonbondedForce', 'HarmonicBondForce', 'HarmonicAngleForce', 'PeriodicTorsionForce', 'PeriodicTorsionForce']
    ct = 0
    thissec = None
    # Why is this a while loop and not a for line in text loop?
    while ct < len(text):
        line = text[ct]
        tmp = line.split()

        # Skip lines starting with comment or which are blank
        if line[0]=='#' or len(tmp) < 1:
            ct+=1
            continue

        # Check first entry to see if it's a section name, if so initialize storage
        if tmp[0] in secnames:
            thissec = tmp[0]
            sections[thissec] = []

        elif tmp[0] in ['DATE','AUTHOR']:
            thissec = tmp[0]
        # Otherwise store
        else:
            if thissec == 'DATE':
                date = line.strip()
            elif thissec == 'AUTHOR':
                author = line.strip()
            else:
                sections[thissec].append(line)

        ct+=1

    # fix date and author in inxml file:
    add_date_and_author(inxml, date, author)
    # Read template force field file
    ff = ForceField(inxml)
    # Use functions to parse sections from target file and add parameters to force field
    param_id_by_section={}
    param_prefix_by_sec = {'NONBON':'n' , 'BOND':'b', 'ANGL':'a', 'DIHE':'t', 'IMPR':'i'}
    env_method_by_sec = {'NONBON': environment.AtomChemicalEnvironment,
            'BOND': environment.BondChemicalEnvironment,
            'ANGL': environment.AngleChemicalEnvironment,
            'DIHE': environment.TorsionChemicalEnvironment,
            'IMPR': environment.ImproperChemicalEnvironment}
    for (idx, name) in enumerate(secnames):
        param_id_by_section[name] = 1
        env_method = env_method_by_sec[name]
        for line in sections[name]:
            # Parse line for parameters
            if name=='NONBON':
                params = _parse_nonbon_line(line)
            elif name=='BOND':
                params = _parse_bond_line(line)
            elif name=='DIHE':
                params = _parse_dihe_line(line)
            elif name=='IMPR':
                params = _parse_impr_line(line)
            elif name=='ANGL':
                params = _parse_angl_line(line)

            # Add parameter ID
            params['id'] = param_prefix_by_sec[name]+str( param_id_by_section[name] )

            smirks = params['smirks']
            #Check smirks is valid for chemical enviroment parsing:
            env = env_method(smirks)

            # If it's not a torsion, just store in straightforward way
            if not (name=='IMPR' or name=='DIHE'):
                # Check for duplicates first
                if ff.getParameter( smirks, force_type = force_section[idx] ):
                    raise ValueError(f"Error: parameter for {smirks} is already present in force field.")
                else:
                    ff.addParameter( params, smirks, force_section[idx], tag[idx] )

                # Increment parameter id
                param_id_by_section[name] +=1
            # If it's a torsion, check to see if there are already parameters and
            # if so, add a new term to this torsion
            else:
                # If we have parameters already
                oldparams = ff.getParameter(smirks, force_type=force_section[idx])
                if oldparams:
                    # Find what number to use
                    idnr = 1
                    paramtag = f"k{idnr}"
                    # This was "while paramtag in params" so it was overwriting k2 etc.
                    while paramtag in oldparams:
                        idnr+=1
                        paramtag = f"k{idnr}"
                    # Construct new param object with updated numbers
                    for paramtag in ('periodicity1', 'phase1', 'idivf1', 'k1'):
                        if paramtag in params:
                            val = params.pop(paramtag)
                            oldparams[paramtag[:-1]+str(idnr) ] = val
                    # Store
                    ff.setParameter( oldparams, smirks=smirks, force_type=force_section[idx])
                else:
                    # Otherwise, just store new parameters
                    ff.addParameter( params, smirks, force_section[idx], tag[idx])
                    # Increment parameter ID
                    param_id_by_section[name] += 1


    # Write SMIRNOFF XML file
    ff.writeFile(outxml)

    # Roundtrip to fix formatting (for some reason etree won't format it properly on first write after modification)
    tmp = ForceField(outxml)
    tmp.writeFile(outxml)


if __name__=="__main__":
    from optparse import OptionParser
    usage_string="""\
    Convert specified SMIRKS-ified AMBER frcmod file into SMIRNOFF FFXML format, inserting converted parameters into a template FFXML file and writing to a new output file.

    usage: convert_frcmod_0.1.py --frcmod test.frcmod --template template.offxml --xml test.offxml
    """
    parser = OptionParser(usage=usage_string)

    parser.add_option('-f', '--frcmod', type = "string", dest='infile', default = None, action="store", help="Name of input smirks-ified frcmod file.")
    parser.add_option('-t', '--template', type="string", dest='inxml', default = None, action ="store", help="Name of template SMIRNOFF offxml file.")
    parser.add_option('-o', '--xml', type="string", dest='outxml', default =None, action="store", help="Name of output SMIRNOFF offxml file.")
    (options,args) = parser.parse_args()

    if (options.infile is None) or (options.inxml is None) or (options.outxml is None):
        parser.print_help()
        parser.error("Input frcmod and template files and output FFXML file must be specified.")


    convert_frcmod_to_ffxml( options.infile, options.inxml, options.outxml )



#!/usr/bin/env python

from forcebalance.molecule import Molecule
from forcebalance.nifty import which
from openeye import oechem
import openmoltools
import os, sys, time, argparse, subprocess

def CalculateMolecularWeight(mol):
    """
    Calculate the molecular weight for an OpenEye molecule.

    Parameters
    ----------
    mol : OEGraphMol

    Returns
    -------
    float
        Molecular weight in g/mol
    """
    result = 0.0
    for atom in mol.GetAtoms():
        elem = atom.GetAtomicNum()
        mass = atom.GetIsotope()
        if (elem != 0 and mass != 0):
            result += oechem.OEGetIsotopicWeight(elem,mass)
        else:
            result += oechem.OEGetAverageWeight(elem)
    return result

def CalculateBoxSize(nmol, molwt, density):
    """
    Calculate the size of a solvent box.

    Parameters
    ----------
    nmol : int
        Number of molecules desired for the box
    molwt : float
        Molecular weight in g/mol
    density : float
        Estimated density in kg/m3 (this should be about 40-50% lower than the real liquid density)

    Returns
    -------
    float
        Length of a cubic solvent box in nm.
    """
    # Calculate total mass of the box in kg
    mass = nmol * molwt / 1000 / 6.022e23
    volume = mass / density
    length = volume**(1./3)/1e-9
    return length

def GenerateBox(pdbin, pdbout, box, nmol, tries):
    """
    Call genbox. (Confirmed working with Gromacs version 4.6.7 and 5.1.4).
    Mainly checks whether genbox ran correctly.

    Parameters
    ----------
    pdbin : str
        Name of input PDB file containing a single molecule.
    pdbout : str
        Name of output PDB file containing solvent box.
    box : float
        Solvent box size, should be determined previously.
    nmol : int
        Number of molecules to go into the solvent box
    tries : int
        Parameter for genbox to try inserting each molecule (tries) times

    Returns
    -------
    None
        If successful, produces "pdbout" containing solvent box.
    """
    if which('gmx'):
        gmxcmd='gmx insert-molecules'
    elif which('genbox'):
        gmxcmd='genbox'
    else:
        raise RuntimeError('gmx and/or genbox not in PATH. Please source Gromacs environment variables.')

    fout=open('genbox.out', 'w')
    ferr=open('genbox.err', 'w')
    subprocess.Popen(f'{gmxcmd} -ci {pdbin} -o genbox.pdb -box {box:.3f} {box:.3f} {box:.3f} -nmol {nmol} -try {tries}',
                     shell=True, stdout=fout, stderr=ferr)
    fout.close()
    ferr.close()
    t0 = time.time()
    print(f"Running {gmxcmd} to create a solvent box...")
    print(f"Time elapsed: {(time.time() - t0): .3f} seconds")
    nmol_out = 0
    for line in open('genbox.err').readlines():
        if 'Output configuration contains' in line:
            nmol_out = int(line.split()[-2])
    if nmol_out == 0:
        raise RuntimeError('genbox failed to produce an output configuration')
    elif nmol_out != nmol:
        raise RuntimeError(f'genbox failed to create a box with {nmol} molecules (actual {nmol_out}); '
                           'please retry with increased box size or number of tries')
    else:
        # genbox throws away the CONECT records in the PDB, this operation adds them back.
        M1 = Molecule(pdbin, build_topology=False)
        M = Molecule('genbox.pdb', build_topology=False)
        solventbox_bonds = []
        # Loop over the number of molecules in the solvent box
        for i in range(nmol):
            for j in M1.bonds:
                # Add the bonds for molecule number "i" in the solvent box
                solventbox_bonds.append((j[0]+i*M1.na, j[1]+i*M1.na))
        M.bonds = solventbox_bonds
        M.write(pdbout)
        print(f"-=# Output #=- Created {pdbout} containing solvent box with {nmol} molecules and length {box:.3f}")

def run_create_mol2_pdb(**kwargs):

    nmol = kwargs['nmol']
    input_txt = kwargs['input']
    resname = kwargs['resname']
    density = kwargs['density']
    tries = kwargs['tries']

    # Disable Gromacs backup file creation
    os.environ['GMX_MAXBACKUP']="-1"

    smiles_string = open(input_txt).readlines()[0].strip()
    print(f"The following SMILES string will be converted: {smiles_string}")

    # create a new molecule
    oemol = oechem.OEGraphMol()
    # convert the SMILES string into a molecule
    if oechem.OESmilesToMol(oemol, smiles_string):
        # do something interesting with mol
        pass
    else:
        print("SMILES string was invalid!")

    # Add explicit
    oechem.OEAddExplicitHydrogens(oemol)

    # Generate a single conformer
    oemol = openmoltools.openeye.generate_conformers(oemol, max_confs=1)

    # Modify residue names
    oechem.OEPerceiveResidues(oemol, oechem.OEPreserveResInfo_All)
    for atom in oemol.GetAtoms():
        thisRes = oechem.OEAtomGetResidue(atom)
        thisRes.SetName(resname)

    # Write output files
    ofs = oechem.oemolostream()
    output_fnms = [f"{resname}.mol2", f'{resname}.pdb']
    for output_fnm in output_fnms:
        if not ofs.open(output_fnm):
            oechem.OEThrow.Fatal(f"Unable to create {output_fnm}")
        oechem.OEWriteMolecule(ofs, oemol)
        print(f"-=# Output #=- Created {output_fnm} containing single molecule")

    grams_per_mole = CalculateMolecularWeight(oemol)

    boxlen = CalculateBoxSize(nmol, grams_per_mole, density)
    GenerateBox(f"{resname}.pdb", f'{resname}-box.pdb', boxlen, nmol, tries)

def main():
    """
    Provide a text file containing a single SMILES string and three-letter residue name.
    Receive (res).pdb and (res).mol2 files containing a single molecule with conformation.
    Receive (res)-box.pdb containing a box with specified number

    Dependencies:
    OpenEye tools (for creating molecule from SMILES)
    openmoltools (for calling OpenEye to generate conformer)
    Gromacs 4.6.7 or 5.1.4 (for calling genbox to create solvent box)
    ForceBalance 1.5.x (for putting information back that was thrown away by genbox)
    """
    parser = argparse.ArgumentParser()
    parser.add_argument('--density', type=int, default=600, help='Specify target density of the solvent box; should be somewhat smaller than true liquid density due to imperfect packing.')
    parser.add_argument('--nmol', type=int, default=256, help='Specify desired number of molecules in the solvent box.')
    parser.add_argument('--tries', type=int, default=10, help='Pass number of tries per molecule to be passed to genbox. Higher = longer runtime but may achieve higher density.')
    parser.add_argument('input', type=str, help='Input file containing a single SMILES string')
    parser.add_argument('resname', type=str, help='Specify a custom residue name for the molecule.')
    print(f"{__file__} called with the following command line:")
    print(' '.join(sys.argv))
    args = parser.parse_args(sys.argv[1:])
    # Create the desired files (.mol2 file containing a single conformation and .pdb file containing solvent box).
    run_create_mol2_pdb(**vars(args))

if __name__ == "__main__":
    main()




import pytest

from openff.toolkit import ForceField


class StaleForceFieldTests:
    @pytest.mark.skip(reason="Needs to be updated for 0.2.0 syntax")
    def test_create_forcefield_from_file_list(self):
        # These offxml files are located in package data path, which is automatically installed and searched
        file_paths = [smirnoff99Frosst_offxml_file_path, tip3p_offxml_file_path]
        # Create a forcefield from multiple offxml files
        ForceField(file_paths)

    @pytest.mark.skip(reason="Needs to be updated for 0.2.0 syntax")
    def test_create_forcefield_from_file_path_iterator(self):
        # These offxml files are located in package data path, which is automatically installed and searched
        file_paths = [smirnoff99Frosst_offxml_file_path, tip3p_offxml_file_path]
        # A generator should work as well
        ForceField(iter(file_paths))

    @pytest.mark.skip(reason="Needs to be updated for 0.2.0 syntax")
    def test_create_gbsa(self):
        """Test reading of ffxml files with GBSA support."""
        ForceField("test_forcefields/Frosst_AlkEthOH_GBSA.offxml")

    @pytest.mark.skip(reason="Needs to be updated for 0.2.0 syntax")
    def test_create_forcefield_from_url(self):
        urls = [
            "https://raw.githubusercontent.com/openforcefield/openff-toolkit/master/openff/toolkit/data/test_forcefields/test_forcefield.offxml",
            "https://raw.githubusercontent.com/openforcefield/openff-toolkit/master/openff/toolkit/data/test_forcefields/tip3p.offxml",
        ]
        # Test creation with smirnoff99frosst URL
        ForceField(urls[0])

    @pytest.mark.skip(reason="Needs to be updated for 0.2.0 syntax")
    def test_create_forcefield_from_url_list(self):
        urls = [
            "https://raw.githubusercontent.com/openforcefield/openff-toolkit/master/openff/toolkit/data/test_forcefields/test_forcefield.offxml",
            "https://raw.githubusercontent.com/openforcefield/openff-toolkit/master/openff/toolkit/data/test_forcefields/tip3p.offxml",
        ]
        # Test creation with multiple URLs
        ForceField(urls)

    @pytest.mark.skip(reason="Needs to be updated for 0.2.0 syntax")
    def test_create_forcefield_from_url_iterator(self):
        urls = [
            "https://raw.githubusercontent.com/openforcefield/openff-toolkit/master/openff/toolkit/data/test_forcefields/test_forcefield.offxml",
            "https://raw.githubusercontent.com/openforcefield/openff-toolkit/master/openff/toolkit/data/test_forcefields/tip3p.offxml",
        ]
        # A generator should work as well
        ForceField(iter(urls))

    @pytest.mark.skip(reason="Needs to be updated for 0.2.0 syntax")
    def test_charge_increment(self):
        """Test parameter assignment using smirnoff99Frosst on laromustine with ChargeIncrementModel."""
        molecules_file_path = get_data_file_path("molecules/laromustine_tripos.mol2")
        molecule = Molecule.from_file(molecules_file_path)
        forcefield = ForceField(["test_forcefields/test_forcefield.offxml", "chargeincrement-test"])
        check_system_creation_from_molecule(forcefield, molecule)
        # TODO: We can't implement a test for chargeincrement yet because we
        #       haven't settled on a SMIRNOFF spec for chargeincrementmodel

    @pytest.mark.skip(reason="Needs to be updated for 0.2.0 syntax")
    def test_create_system_molecules_parmatfrosst_gbsa(self):
        """Test creation of a System object from small molecules to test parm@frosst force field with GBSA support."""
        molecules_file_path = get_data_file_path("molecules/AlkEthOH_test_filt1_tripos.mol2")
        check_parameter_assignment(
            offxml_file_path="test_forcefields/Frosst_AlkEthOH_GBSA.offxml",
            molecules_file_path=molecules_file_path,
        )
        # TODO: Figure out if we just want to check that energy is finite (this is what the original test did,
        #       or compare numerically to a reference system.

    @pytest.mark.skip(reason="Needs to be updated for 0.2.0 syntax")
    def test_deep_copy(self):
        force_field = ForceField(smirnoff99Frosst_offxml_file_path)
        # Deep copy
        force_field2 = copy.deepcopy(force_field)
        assert_forcefields_equal(
            force_field,
            force_field2,
            "ForceField deep copy does not match original ForceField",
        )

    @pytest.mark.skip(reason="Needs to be updated for 0.2.0 syntax")
    # TODO: This should check the output of forcefield.to_dict
    def test_serialize(self):
        force_field = ForceField(smirnoff99Frosst_offxml_file_path)
        # Serialize/deserialize
        serialized_forcefield = force_field.__getstate__()
        force_field2 = ForceField.__setstate__(serialized_forcefield)
        assert_forcefields_equal(
            force_field,
            force_field2,
            "Deserialized serialized ForceField does not match original ForceField",
        )


@pytest.mark.skip(reason="Needs to be updated for 0.2.0 syntax")
def test_electrostatics_options(self):
    """Test parameter assignment using smirnoff99Frosst on laromustine with various long-range electrostatics options."""
    from functools import partial

    molecules_file_path = get_data_file_path("molecules/laromustine_tripos.mol2")
    molecule = Molecule.from_file(molecules_file_path)
    forcefield = ForceField([smirnoff99Frosst_offxml_file_path, charge_increment_offxml_file_path])
    for method in ["PME", "reaction-field", "Coulomb"]:
        # Change electrostatics method
        forcefield.forces["Electrostatics"].method = method
        f = partial(check_system_creation_from_molecule, forcefield, molecule)
        f.description = f"Testing {offxml_file_path} parameter assignment using molecule {molecule.name}"
        # yield f
    # TODO: Implement a similar test, where we compare OpenMM energy evals from an
    #       AMBER-parameterized system to OFF-parameterized systems


class StaleIOTests:
    @pytest.mark.skip(reason="Needs to be updated for 1.0.0 syntax")
    def test_to_xml(self):
        forcefield = ForceField(smirnoff99Frosst_offxml_filename)
        # Retrieve XML as a string
        xml = forcefield.to_xml()
        # Restore ForceField from XML
        forcefield2 = ForceField(xml)
        assert_forcefields_equal(
            cls.forcefield,
            forcefield2,
            "ForceField serialized to XML does not match original ForceField",
        )

    # TODO: Remove ForceField from this whole file. All tests should be for converting between hierarchical SMIRNOFF
    #       dicts and XML
    @pytest.mark.skip(reason="Needs to be updated for 1.0.0 syntax")
    def test_save(self):
        """Test writing and reading of SMIRNOFF in XML format."""
        forcefield = ForceField(smirnoff99Frosst_offxml_filename)
        # Write XML to a file
        with TemporaryDirectory() as tmpdir:
            offxml_tmpfile = os.path.join(tmpdir, "forcefield.offxml")
            forcefield.save(offxml_tmpfile)
            forcefield2 = ForceField(offxml_tmpfile)
            assert_forcefields_equal(
                cls.forcefield,
                forcefield2,
                "ForceField written to .offxml does not match original ForceField",
            )



"""
TypedMolecule TODOs
* Testing
* Detemine whether atomic numbers are required, or whether a TypedMol could have a mix of
  TypedAtom and TypedParticles, where the latter doesn't need atomic nubmers
* Determine whether typedmols should have isomorphism operations
* Should be able to go to hill formula (?), networkx, openmm
* Topology serialization will have trouble here - Won't know whether it's trying to
  deserialize a Molecule or a TypedMolecule.

"""

import functools
from collections.abc import Generator, Iterable
from typing import TYPE_CHECKING, NoReturn, Union

from openff.units.elements import MASSES, SYMBOLS

from openff.toolkit import unit
from openff.toolkit.topology.molecule import (
    AtomMetadataDict,
    HierarchyElement,
    HierarchyScheme,
    Molecule,
    _atom_nums_to_hill_formula,
    _has_unique_atom_names,
)
from openff.toolkit.utils.exceptions import UnsupportedMoleculeConversionError
from openff.toolkit.utils.utils import deserialize_numpy, serialize_numpy

if TYPE_CHECKING:
    import networkx as nx

    from openff.toolkit import Quantity, Topology
    from openff.toolkit.topology.molecule import FrozenMolecule


class _SimpleMolecule:
    def __init__(self):
        self.atoms = []
        self.bonds = []
        self._conformers = None
        self._hierarchy_schemes = dict()

    def add_atom(self, atomic_number: int, **kwargs):
        atom = _SimpleAtom(atomic_number, self, **kwargs)
        self.atoms.append(atom)

    def add_bond(self, atom1, atom2, **kwargs):
        if isinstance(atom1, int) and isinstance(atom2, int):
            atom1_atom = self.atoms[atom1]
            atom2_atom = self.atoms[atom2]
        elif isinstance(atom1, _SimpleAtom) and isinstance(atom2, _SimpleAtom):
            atom1_atom = atom1
            atom2_atom = atom2
        else:
            raise ValueError(
                "Invalid inputs to molecule._add_bond. Expected ints or Atoms. "
                f"Received {atom1} (type {type(atom1)}) and {atom2} (type {type(atom2)}) "
            )
        bond = _SimpleBond(atom1_atom, atom2_atom, **kwargs)
        self.bonds.append(bond)

    @property
    def conformers(self) -> list["Quantity"] | None:
        return self._conformers

    def add_conformer(self, conformer):
        if self._conformers is None:
            self._conformers = list()
        self._conformers.append(conformer)

    @property
    def n_atoms(self) -> int:
        return len(self.atoms)

    @property
    def n_bonds(self) -> int:
        return len(self.bonds)

    @property
    def n_conformers(self) -> int:
        return 0 if self._conformers is None else len(self._conformers)

    def atom(self, index: int) -> "_SimpleAtom":
        return self.atoms[index]

    def atom_index(self, atom) -> int:
        return self.atoms.index(atom)

    def bond(self, index: int) -> "_SimpleBond":
        return self.bonds[index]

    def get_bond_between(self, atom1_index, atom2_index):
        atom1 = self.atom(atom1_index)
        atom2 = self.atom(atom2_index)
        for bond in atom1.bonds:
            for atom in bond.atoms:
                if atom is atom2:
                    return bond

    @property
    def angles(
        self,
    ) -> Generator[
        tuple[
            "_SimpleAtom",
            "_SimpleAtom",
            "_SimpleAtom",
        ],
        None,
        None,
    ]:
        for atom1 in self.atoms:
            for atom2 in atom1.bonded_atoms:
                for atom3 in atom2.bonded_atoms:
                    if atom1 is atom3:
                        continue
                    if atom1.molecule_atom_index < atom3.molecule_atom_index:
                        yield (atom1, atom2, atom3)
                    else:
                        # Do not return i.e. (2, 1, 0), only (0, 1, 2)
                        pass

    @property
    def propers(
        self,
    ) -> Generator[
        tuple[
            "_SimpleAtom",
            "_SimpleAtom",
            "_SimpleAtom",
            "_SimpleAtom",
        ],
        None,
        None,
    ]:
        for atom1 in self.atoms:
            for atom2 in atom1.bonded_atoms:
                for atom3 in atom2.bonded_atoms:
                    if atom1 is atom3:
                        continue
                    for atom4 in atom3.bonded_atoms:
                        if atom4 in (atom1, atom2):
                            continue

                        if atom1.molecule_atom_index < atom4.molecule_atom_index:
                            yield (atom1, atom2, atom3, atom4)
                        else:
                            # Do no duplicate
                            pass

    @property
    def impropers(
        self,
    ) -> Generator[
        tuple[
            "_SimpleAtom",
            "_SimpleAtom",
            "_SimpleAtom",
            "_SimpleAtom",
        ],
        None,
        None,
    ]:
        for atom1 in self.atoms:
            for atom2 in atom1.bonded_atoms:
                for atom3 in atom2.bonded_atoms:
                    if atom1 is atom3:
                        continue
                    for atom3i in atom2.bonded_atoms:
                        if atom3i == atom3:
                            continue
                        if atom3i == atom1:
                            continue

                        yield (atom1, atom2, atom3, atom3i)

    @property
    def smirnoff_impropers(
        self,
    ) -> Generator[
        tuple[
            "_SimpleAtom",
            "_SimpleAtom",
            "_SimpleAtom",
            "_SimpleAtom",
        ],
        None,
        None,
    ]:
        for improper in self.impropers:
            if len(list(improper[1].bonded_atoms)) == 3:
                yield improper

    @property
    def amber_impropers(
        self,
    ) -> Generator[
        tuple[
            "_SimpleAtom",
            "_SimpleAtom",
            "_SimpleAtom",
            "_SimpleAtom",
        ],
        None,
        None,
    ]:
        for improper in self.smirnoff_impropers:
            yield (improper[1], improper[0], improper[2], improper[3])

    @property
    def n_angles(self) -> int:
        return len(list(self.angles))

    @property
    def n_propers(self) -> int:
        return len(list(self.propers))

    @property
    def n_impropers(self) -> int:
        return len(list(self.impropers))

    @property
    def hill_formula(self) -> str:
        """
        Return the Hill formula of this molecule.

        If any atoms are non-elemental, defined by having an atomic number of 0,
        return INVALID instead.
        """
        return self.to_hill_formula()

    @property
    def hierarchy_schemes(self) -> dict[str, "HierarchyScheme"]:
        return self._hierarchy_schemes

    def to_hill_formula(self) -> str:
        atom_nums: list[int] = [atom.atomic_number for atom in self.atoms]

        if min(atom_nums) < 0:
            return "INVALID"

        return _atom_nums_to_hill_formula(atom_nums)

    def to_networkx(self) -> "nx.Graph":
        # TODO: Custom attribtues should probably be attached to the nodes (and possibly also
        #       the edges?). See for more:
        #       https://github.com/openforcefield/openff-toolkit/pull/1179#discussion_r808549385
        import networkx as nx

        graph: nx.classes.graph.Graph = nx.Graph()

        for atom_index, atom in enumerate(self.atoms):
            graph.add_node(
                atom_index,
                atomic_number=atom.atomic_number,
            )
        for bond in self.bonds:
            graph.add_edge(
                bond.atom1_index,
                bond.atom2_index,
            )

        return graph

    def to_topology(self) -> "Topology":
        from openff.toolkit.topology import Topology

        return Topology.from_molecules([self])

    def nth_degree_neighbors(self, n_degrees):
        from openff.toolkit.topology.molecule import (
            _nth_degree_neighbors_from_graphlike,
        )

        return _nth_degree_neighbors_from_graphlike(graphlike=self, n_degrees=n_degrees)

    @classmethod
    def _from_subgraph(cls, subgraph: "nx.Graph"):
        molecule = cls()

        # The subgraph stores indices that might not start at zero (i.e. topology indices)
        # but we need 0-index indices (like molecule indices) to add bonds
        offset = min(subgraph.nodes())

        for _, node_data in subgraph.nodes(data=True):
            # Hierarchy metadata like residue name needs to be passed in as a separate argument,
            # so we extract those values from the node data
            metadata = dict()
            for key, val in node_data.items():
                if key in [
                    "residue_name",
                    "residue_number",
                    "insertion_code",
                    "chain_id",
                ]:
                    metadata[key] = val
            # Then we remove the metadata items that we took from the node data
            for key in metadata.keys():
                del node_data[key]

            molecule.add_atom(metadata=metadata, **node_data)

        for topology_index1, topology_index2, _edge_data in subgraph.edges(data=True):
            molecule.add_bond(topology_index1 - offset, topology_index2 - offset)

        return molecule

    # TODO: Implement me - shouldn't need to be too different than Molecule.add_hierarchy_scheme
    #       since the extra chemical information is unrelated to hierarchy/metadata
    def add_hierarchy_scheme(
        self,
        uniqueness_criteria: Iterable[str],
        iterator_name: str,
    ) -> NoReturn:
        raise NotImplementedError()

    def to_dict(self) -> dict:
        molecule_dict = dict()
        special_serialization_logic = [
            "atoms",
            "bonds",
            "conformers",
            "hierarchy_schemes",
            "molecule",
        ]
        for attr_name, attr_val in self.__dict__.items():
            if attr_name.startswith("_"):
                continue
            if attr_name in special_serialization_logic:
                continue
            molecule_dict[attr_name] = attr_val

        molecule_dict["atoms"] = [atom.to_dict() for atom in self.atoms]

        molecule_dict["bonds"] = [bond.to_dict() for bond in self.bonds]

        if self._conformers is None:
            molecule_dict["conformers"] = None
        else:
            molecule_dict["conformers"] = []
            molecule_dict["conformers_unit"] = "angstrom"  # Have this defined as a class variable?
            for conf in self._conformers:
                conf_unitless = conf.m_as(unit.angstrom)
                conf_serialized, _ = serialize_numpy(conf_unitless)
                molecule_dict["conformers"].append(conf_serialized)

        molecule_dict["hierarchy_schemes"] = dict()
        for iter_name, hier_scheme in self.hierarchy_schemes.items():
            molecule_dict["hierarchy_schemes"][iter_name] = hier_scheme.to_dict()

        return molecule_dict

    @classmethod
    def from_dict(cls, molecule_dict):
        molecule = cls()

        atom_dicts = molecule_dict.pop("atoms")
        for atom_dict in atom_dicts:
            molecule.atoms.append(_SimpleAtom.from_dict(atom_dict))
            molecule.atoms[-1]._molecule = molecule

        bond_dicts = molecule_dict.pop("bonds")

        for bond_dict in bond_dicts:
            atom1_index = bond_dict["atom1_index"]
            atom2_index = bond_dict["atom2_index"]
            molecule.add_bond(atom1=molecule.atom(atom1_index), atom2=molecule.atom(atom2_index))

        conformers = molecule_dict.pop("conformers")
        if conformers is None:
            molecule._conformers = None
        else:
            conformers_unit = molecule_dict.pop("conformers_unit")
            molecule._conformers = list()
            for ser_conf in conformers:
                conformers_shape = (molecule.n_atoms, 3)
                conformer_unitless = deserialize_numpy(ser_conf, conformers_shape)
                conformer = unit.Quantity(conformer_unitless, conformers_unit)
                molecule._conformers.append(conformer)

        hier_scheme_dicts = molecule_dict.pop("hierarchy_schemes")
        for iterator_name, hierarchy_scheme_dict in hier_scheme_dicts.items():
            molecule._hierarchy_schemes[iterator_name] = HierarchyScheme(
                parent=molecule,
                uniqueness_criteria=tuple(hierarchy_scheme_dict["uniqueness_criteria"]),
                iterator_name=iterator_name,
            )

            for element_dict in hierarchy_scheme_dict["hierarchy_elements"]:
                molecule._hierarchy_schemes[iterator_name].add_hierarchy_element(
                    identifier=element_dict["identifier"],
                    atom_indices=element_dict["atom_indices"],
                )

        for key, val in molecule_dict.items():
            setattr(molecule, key, val)

        return molecule

    @classmethod
    def from_molecule(cls, molecule: Molecule):
        """Generate an MM molecule from an OpenFF Molecule."""
        mm_molecule = cls()
        for atom in molecule.atoms:
            mm_molecule.add_atom(
                atomic_number=atom.atomic_number,
                metadata=atom.metadata,
            )

        for bond in molecule.bonds:
            mm_molecule.add_bond(
                atom1=mm_molecule.atom(bond.atom1_index),
                atom2=mm_molecule.atom(bond.atom2_index),
            )

        if molecule.conformers is not None:
            for conformer in molecule.conformers:
                mm_molecule.add_conformer(conformer)

        for name, hierarchy_scheme in molecule.hierarchy_schemes.items():
            assert name == hierarchy_scheme.iterator_name
            mm_molecule.add_hierarchy_scheme(
                uniqueness_criteria=hierarchy_scheme.uniqueness_criteria,
                iterator_name=hierarchy_scheme.iterator_name,
            )

        return mm_molecule

    def to_molecule(self) -> NoReturn:
        raise UnsupportedMoleculeConversionError(
            "The information content of a _SimpleMolecule is insufficient for creating "
            "an OpenFF Molecule with sufficiently specified chemistry."
        )

    def is_isomorphic_with(self, other: Union["FrozenMolecule", "_SimpleMolecule", "nx.Graph"], **kwargs) -> bool:
        """
        Check for pseudo-isomorphism.

        This currently checks that the two molecules have
        * The same number of atoms
        * The same number of bonds
        * Atoms with matching atomic numbers, in order

        This currently does NOT checks that the two molecules have
        * Topologically identical bond graphs
        """
        return _SimpleMolecule.are_isomorphic(
            self,
            other,
            return_atom_map=False,
        )[0]

    @staticmethod
    def are_isomorphic(
        mol1: Union["FrozenMolecule", "_SimpleMolecule", "nx.Graph"],
        mol2: Union["FrozenMolecule", "_SimpleMolecule", "nx.Graph"],
        return_atom_map: bool = False,
    ) -> tuple[bool, dict[int, int] | None]:
        import networkx

        _cls = _SimpleMolecule

        if isinstance(mol1, networkx.Graph) and isinstance(mol2, networkx.Graph):
            graph1 = _SimpleMolecule._from_subgraph(mol1).to_networkx()
            graph2 = _SimpleMolecule._from_subgraph(mol2).to_networkx()

        elif isinstance(mol1, networkx.Graph):
            assert isinstance(mol2, _cls)
            graph1 = _SimpleMolecule._from_subgraph(mol1).to_networkx()
            graph2 = mol2.to_networkx()

        elif isinstance(mol2, networkx.Graph):
            assert isinstance(mol1, _cls)
            graph1 = mol1.to_networkx()
            graph2 = _SimpleMolecule._from_subgraph(mol2).to_networkx()

        else:
            # static methods (by definition) know nothing about their class,
            # so the class to compare to must be hard-coded here
            if not (isinstance(mol1, _cls) and isinstance(mol2, _cls)):
                return False, None

            graph1 = mol1.to_networkx()
            graph2 = mol2.to_networkx()

        def node_match_func(node1, node2):
            return node1["atomic_number"] == node2["atomic_number"]

        edge_match_func = None

        matcher = networkx.algorithms.isomorphism.GraphMatcher(
            graph1, graph2, node_match=node_match_func, edge_match=edge_match_func
        )

        if matcher.is_isomorphic():
            if return_atom_map:
                topology_atom_map = matcher.mapping

                return True, {key: topology_atom_map[key] for key in sorted(topology_atom_map)}

            else:
                return True, None
        else:
            return False, None

    def generate_unique_atom_names(self):
        """Generate unique atom names. See `Molecule.generate_unique_atom_names`."""
        from collections import defaultdict

        element_counts: defaultdict[str, int] = defaultdict(int)
        for atom in self.atoms:
            symbol = SYMBOLS[atom.atomic_number]
            element_counts[symbol] += 1

            atom.name = symbol + str(element_counts[symbol]) + "x"

    @property
    def has_unique_atom_names(self) -> bool:
        """``True`` if the molecule has unique atom names, ``False`` otherwise."""
        return _has_unique_atom_names(self)

    def __getattr__(self, name: str) -> list["HierarchyElement"]:
        """If a requested attribute is not found, check the hierarchy schemes"""
        try:
            return self.__dict__["_hierarchy_schemes"][name].hierarchy_elements
        except KeyError:
            raise AttributeError(f"'{self.__class__.__name__}' object has no attribute {name!r}")

    def __deepcopy__(self, memo):
        return self.__class__.from_dict(self.to_dict())


_SimpleMolecule.add_hierarchy_scheme = Molecule.add_hierarchy_scheme  # type: ignore[assignment]
_SimpleMolecule.update_hierarchy_schemes = Molecule.update_hierarchy_schemes  # type: ignore[attr-defined]


class _SimpleAtom:
    def __init__(
        self,
        atomic_number: int,
        molecule: _SimpleMolecule | None = None,
        metadata: AtomMetadataDict | None = None,
        name: str = "",
        **kwargs,
    ):
        if metadata is None:
            self.metadata = AtomMetadataDict()
        else:
            self.metadata = AtomMetadataDict(metadata)

        # This _could_ be hidden away into _metadata
        self._name = name
        self._atomic_number = atomic_number
        self._molecule = molecule
        self._bonds: list[_SimpleBond] = list()
        for key, val in kwargs.items():
            setattr(self, key, val)

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, value: str):
        self._name = value

    @property
    def atomic_number(self) -> int:
        return self._atomic_number

    @atomic_number.setter
    def atomic_number(self, value):
        if not isinstance(value, int):
            raise ValueError("atomic_number must be an integer")
        if value < 0:
            raise ValueError("atomic_number must be non-negative. An atomic number of 0 is acceptable.")
        self._atomic_number = value

    @property
    def symbol(self) -> str:
        return SYMBOLS[self.atomic_number]

    @property
    def mass(self) -> "Quantity":
        return MASSES[self.atomic_number]

    @property
    def molecule(self):
        """The ``Molecule`` this atom is part of."""
        return self._molecule

    @property
    def bonds(self):
        return self._bonds

    def add_bond(self, bond):
        self._bonds.append(bond)

    @property
    def bonded_atoms(self):
        for bond in self._bonds:
            for atom in [bond.atom1, bond.atom2]:
                if atom is not self:
                    yield atom

    @functools.cached_property
    def molecule_atom_index(self) -> int:
        return self.molecule.atoms.index(self)

    def to_dict(self) -> dict[str, dict | str | int]:
        atom_dict: dict[str, dict | str | int] = dict()
        atom_dict["metadata"] = dict(self.metadata)
        atom_dict["atomic_number"] = self._atomic_number
        atom_dict["name"] = self._name

        keys_to_skip = ["metadata", "molecule", "bonds"]

        for attr_name, attr_val in self.__dict__.items():
            if attr_name.startswith("_"):
                continue
            if attr_name in keys_to_skip:
                continue
            atom_dict[attr_name] = attr_val

        return atom_dict

    @classmethod
    def from_dict(cls, atom_dict: dict):
        atom = cls(**atom_dict)
        return atom


class _SimpleBond:
    def __init__(self, atom1, atom2, **kwargs):
        self.molecule = atom1.molecule
        self.atom1 = atom1
        self.atom2 = atom2

        atom1.add_bond(self)
        atom2.add_bond(self)

        for key, val in kwargs.items():
            setattr(self, key, val)

    @property
    def atoms(self) -> list[_SimpleAtom]:
        return [self.atom1, self.atom2]

    @property
    def atom1_index(self) -> int:
        return self.atom1.molecule_atom_index

    @property
    def atom2_index(self) -> int:
        return self.atom2.molecule_atom_index

    def to_dict(self) -> dict[str, int]:
        return {
            "atom1_index": self.atom1.molecule_atom_index,
            "atom2_index": self.atom2.molecule_atom_index,
        }



#!/bin/env python

import numpy as np
from openff.toolkit.typing.engines.smirnoff.forcefield import ForceField

# Function definitions for parsing sections within parameter file
def _parse_nonbon_line( line ):
    """Parse an AMBER frcmod nonbon line and return relevant parameters in a dictionary. AMBER uses rmin_half and epsilon in angstroms and kilocalories per mole."""
    tmp = line.split()
    params = {}
    params['smirks'] = tmp[0]
    params['rmin_half'] = tmp[1]
    params['epsilon'] = tmp[2]

    return params

def _parse_bond_line( line ):
    """Parse an AMBER frcmod BOND line and return relevant parameters in a dictionary. AMBER uses length and force constant, with the factor of two dropped. Here we multiply by the factor of two before returning. Units are angstroms and kilocalories per mole per square angstrom."""

    tmp = line.split()
    params = {}
    params['smirks'] = tmp[0]
    params['k'] = str(2*float(tmp[1]))
    params['length'] = tmp[2]
    return params

def _parse_angl_line( line ):
    """Parse an AMBER frcmod ANGL line and return relevant parameters in a dictionary. AMBER uses angle and force constant, with the factor of two dropped. Here we multiply by the factor of two before returning. Units are degrees and kilocalories per mole."""

    tmp = line.split()
    params = {}
    params['smirks'] = tmp[0]
    params['k'] = str(2*float(tmp[1]))
    params['angle'] = tmp[2]
    return params

def _parse_dihe_line( line ):
    """Parse an AMBER frcmod DIHE line and return relevant parameters in a dictionary. Units for k are kilocalories per mole."""

    tmp = line.split()
    params = {}
    params['smirks'] = tmp[0]
    params['idivf1'] = tmp[1]
    params['k1'] = tmp[2]
    params['phase1'] = tmp[3]
    params['periodicity1'] = str(int(np.abs(float(tmp[4]))))
    return params

def _parse_impr_line( line ):
    """Parse an AMBER frcmod DIHE line and return relevant parameters in a dictionary. Units for k are kilocalories per mole."""

    tmp = line.split()
    params = {}
    params['smirks'] = tmp[0]
    params['k1'] = tmp[1]
    params['phase1'] = tmp[2]
    params['periodicity1'] = str(int(np.abs(float(tmp[3]))))
    return params


def parse_frcmod(frcmod_file_name):
    """
    Parse a smirnoffish FRCMOD file, returning a dict of lists containing the individual parameters that would go
    into a "SMIRNOFF data" dictionary. When combined with section headers, this dictionary is suitable to be fed
    into the ForceField initializer

    Parameters
    ----------
    frcmod_file_name : str
        path to a smirnoffish FRCMOD file

    Returns
    -------
    parameter_lists : dict of list
        Hierarchical dict of lists, containing parameter entries compliant with the SMIRNOFF spec version 0.2
    author : str or None
        Contents of the Author field, if any
    date: str or None
        Contents of the Date field, if any

    """

    # Obtain sections from target file
    file = open(frcmod_file_name, 'r')
    text = file.readlines()
    file.close()
    sections = {}
    # Section names from frcmod which we will parse
    sec_names = ['NONBON', 'BOND', 'ANGL', 'IMPR', 'DIHE']
    sec_name_2_param_prefix = {'NONBON':'n' , 'BOND':'b', 'ANGL':'a', 'DIHE':'t', 'IMPR':'i'}
    sec_name_2_param_index = dict((sn, 1) for sn in sec_names)
    # Tags that will be used in the FFXML for these (same order)
    force_tags = ['Atom', 'Bond', 'Angle', 'Improper', 'Proper']
    # Force names in the FFXML (same order)
    force_sections = ['vdW', 'Bonds', 'Angles', 'ImproperTorsions', 'ProperTorsions']
    sec_name_2_force_tag = dict((sn, t) for sn, t in zip(sec_names, force_tags))
    sec_name_2_force_section = dict((sn, fs) for sn, fs in zip(sec_names, force_sections))

    date = None
    author = None
    for line in text:
    #while ct < len(text):
        #line = text[ct]
        linesp = line.split()

        # Skip lines starting with comment or which are blank
        if line[0]=='#' or len(linesp) < 1:
            continue

        # Check first entry to see if it's a section name, if so initialize storage
        if linesp[0] in sec_names:
            this_sec = linesp[0]
            sections[this_sec] = []

        elif linesp[0] in ['DATE', 'AUTHOR']:
            this_sec = linesp[0]
        # Otherwise store
        else:
            if this_sec == 'DATE':
                date = line.strip()
            elif this_sec == 'AUTHOR':
                author = line.strip()
            else:
                sections[this_sec].append(line)


    parameter_lists = {}

    # Use functions to parse sections from target file and add parameters to force field
    for (sec_idx, sec_name) in enumerate(sec_names):
        param_list = []
        # sections[sec_name] = []
        for line in sections[sec_name]:
            # Parse line for parameters
            if sec_name=='NONBON':
                param = _parse_nonbon_line(line)
            elif sec_name=='BOND':
                param = _parse_bond_line(line)
            elif sec_name=='DIHE':
                param = _parse_dihe_line(line)
            elif sec_name=='IMPR':
                param = _parse_impr_line(line)
            elif sec_name=='ANGL':
                param = _parse_angl_line(line)

            # Add parameter ID
            param_prefix = sec_name_2_param_prefix[sec_name]
            param_index = sec_name_2_param_index[sec_name]
            param['id'] = param_prefix + str( param_index )

            # If we're dealing with a simple parameter, just append it to the list
            if not (sec_name in ['IMPR', 'DIHE']):
                param_list.append(param)
                # Increment parameter index
                sec_name_2_param_index[sec_name] = sec_name_2_param_index[sec_name] + 1

            # If we're dealing with a potentially multi-term parameter, check if this SMIRKS is already in the list
            else:
                param_matched = False
                for existing_param in param_list:
                    if existing_param['smirks'] != param['smirks']:
                        continue
                    param_matched = True
                    # Find the lowest unoccupied torsion term index
                    term_index = 1
                    while f'k{term_index}' in existing_param:
                        term_index += 1
                    existing_param[f'k{term_index}'] = param['k1']
                    existing_param[f'phase{term_index}'] = param['phase1']
                    existing_param[f'periodicity{term_index}'] = param['periodicity1']
                    existing_param[f'idivf{term_index}'] = param['idivf1']
                    break

                # If the SMIRKS isn't already known, initialize this as a new parameter, since it could be the
                # first term of a multiterm torsion
                if not(param_matched):
                    param_list.append(param)
                    # Increment parameter index ONLY for the first term of a torsion that's found.
                    sec_name_2_param_index[sec_name] = sec_name_2_param_index[sec_name] + 1

        force_section = sec_name_2_force_section[sec_name]
        force_tag = sec_name_2_force_tag[sec_name]
        parameter_lists[force_section] = {force_tag: param_list.copy()}

    return parameter_lists, author, date

# Main conversion functionality
def convert_frcmod_to_ffxml( infile, inxml, outxml ):
    """Convert a modified AMBER frcmod (with SMIRKS replacing atom types) to SMIRNOFF ffxml format by inserting parameters into a template ffxml file.

    Parameters
    ----------
    infile : str
        File name of input SMIRKS-ified frcmod file containing parameters
    inxml : str
        File name of template SMIRNOFF FFXML file into which to insert these parameters.
    outxml : str
        File name of resulting output SMIRNOFF FFXML

    Notes:
    -------
    Input XML file will normally be the template of a SMIRNOFF XML file without any parameters present (but with requisite force types already specified).
    """
    from openff.toolkit.typing.engines.smirnoff import XMLParameterIOHandler



    io_handler = XMLParameterIOHandler()
    smirnoff_data = io_handler.parse_file(inxml)
    parameter_lists, author, date = parse_frcmod(infile)
    for section_tag in parameter_lists:
        # Nest the actual parameter lists one level deeper, by their parameter_list_tag (eg Bonds > Bond > List)
        parameter_list_tag = list(parameter_lists[section_tag].keys())[0]
        smirnoff_data['SMIRNOFF'][section_tag][parameter_list_tag] = parameter_lists[section_tag][parameter_list_tag]
    ff=ForceField()
    ff._load_smirnoff_data(smirnoff_data)
    print(ff.to_string())

    # TODO: Add author and date

    ff.to_file(outxml, io_format='XML')
    # # Write SMIRNOFF XML file
    # ff.writeFile(outxml)
    #
    # # Roundtrip to fix formatting (for some reason etree won't format it properly on first write after modification)
    # tmp = ForceField(outxml)
    # tmp.writeFile(outxml)


if __name__=="__main__":
    from optparse import OptionParser
    usage_string="""\
    Convert specified SMIRKS-ified AMBER frcmod file into SMIRNOFF FFXML format, inserting converted parameters into a template FFXML file and writing to a new output file.

    usage: convert_frcmod_0.2.py --frcmod test.frcmod --template template.offxml --xml test.offxml
    """
    parser = OptionParser(usage=usage_string)

    parser.add_option('-f', '--frcmod', type = "string", dest='infile', default = None, action="store", help="Name of input smirks-ified frcmod file.")
    parser.add_option('-t', '--template', type="string", dest='inxml', default = None, action ="store", help="Name of template SMIRNOFF offxml file.")
    parser.add_option('-o', '--xml', type="string", dest='outxml', default =None, action="store", help="Name of output SMIRNOFF offxml file.")
    (options,args) = parser.parse_args()

    if (options.infile is None) or (options.inxml is None) or (options.outxml is None):
        parser.print_help()
        parser.error("Input frcmod and template files and output FFXML file must be specified.")


    convert_frcmod_to_ffxml( options.infile, options.inxml, options.outxml )



#!/bin/env python

"""
Illustrate how to combine a SMIRNOFF parameterized small molecule with an AMBER parameterized protein using ParmEd.

"""

#
# Load and parameterize the small molecule
#

# Load the small molecule
from openff.toolkit.utils import get_data_file_path
ligand_filename = get_data_file_path('molecules/toluene.mol2')
molecule = Molecule.from_file(ligand_filename)

# Load the smirnoff99Frosst force field
from openff.toolkit.typing.engines import smirnoff
forcefield = smirnoff.ForceField('test_forcefields/test_forcefield.offxml')

# Create a ParmEd structure for the molecule
molecule_structure = forcefield.create_parmed_structure(topology=molecule.to_topology(), positions=molecule.positions)
print('Molecule:', molecule_structure)

#
# Load and parameterize the protein
#

# Load the protein topology
protein_pdb_filename = get_data_file_path('proteins/T4-protein.pdb')
protein_pdbfile = app.PDBFile(protein_pdb_filename)

# Load the AMBER protein force field, along with a solvent force field
from simtk.openmm import app
protein_forcefield = 'amber99sbildn.xml'
solvent_forcefield = 'tip3p.xml'
forcefield = app.ForceField(protein_forcefield, solvent_forcefield)

# Parameterize the protein
protein_system = forcefield.createSystem(proteinpdb.topology)

# Create a ParmEd Structure for the protein
protein_structure = parmed.openmm.load_topology(proteinpdb.topology,
                                                protein_system,
                                                xyz=proteinpdb.positions)
print('Protein:', protein_structure)

# Combine the ParmEd Structure objects to produce a fully parameterized complex
# that can now be exported to AMBER, CHARMM, OpenMM, and other packages
# Note that the addition should always add the small molecule second so the box vectors if the first item (the protein) are to be preserved
complex_structure = protein_structure + molecule_structure
print('Complex:', complex_structure)

# TODO: How can we add solvent while ensuring the ligand doesn't overlap with solvent molecules?
# TODO: Can we have SMIRNOFF ForceField create an OpenMM ffxml file for the ligand, and then use the OpenMM pipeline?
# TODO: Or can OpenMM just use dummy parameters?
